/* CYGNUS LOCAL -- amylaar/calmrisc16 */
;; CalmRISC16 Machine description
;; Copyright (C) 1997, 1998, 1999, 2000 Free Software Foundation, Inc.
;; Contributed by Red Hat

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;- See file "rtl.def" for documentation on define_insn, match_*, et. al.

(define_constants [
   (R0_REG 0)
   (R1_REG 1)
   (R2_REG 2)
   (R3_REG 3)
   (R4_REG 4)
   (R5_REG 5)
   (R6_REG 6)
   (R7_REG 7)
   (R8_REG 8)
   (R10_REG 10)
   (A8_REG 8)
   (A9_REG 9)
   (A10_REG 10)
   (LINK_REG 14)
   (SP_REG 15)
   (T_REG 16)
   (Z0_REG 17)
   (Z1_REG 18)
])


;; ::::::::::::::::::::
;; ::
;; :: Constraints
;; ::
;; ::::::::::::::::::::

;; Standard Constraints
;;
;; `m' A memory operand is allowed, with any kind of address that the
;;     machine supports in general.
;;
;; `o' A memory operand is allowed, but only if the address is
;;     "offsettable".  This means that adding a small integer (actually, the
;;     width in bytes of the operand, as determined by its machine mode) may be
;;     added to the address and the result is also a valid memory address.
;;
;; `V' A memory operand that is not offsettable.  In other words,
;;     anything that would fit the `m' constraint but not the `o' constraint.
;;
;; `<' A memory operand with autodecrement addressing (either
;;     predecrement or postdecrement) is allowed.
;;
;; `>' A memory operand with autoincrement addressing (either
;;     preincrement or postincrement) is allowed.
;;
;; `r' A register operand is allowed provided that it is in a general
;;     register.
;;
;; `d', `a', `f', ...
;;     Other letters can be defined in machine-dependent fashion to stand for
;;     particular classes of registers.  `d', `a' and `f' are defined on the
;;     68000/68020 to stand for data, address and floating point registers.
;;
;; `i' An immediate integer operand (one with constant value) is allowed.
;;     This includes symbolic constants whose values will be known only at
;;     assembly time.
;;
;; `n' An immediate integer operand with a known numeric value is allowed.
;;     Many systems cannot support assembly-time constants for operands less
;;     than a word wide.  Constraints for these operands should use `n' rather
;;     than `i'.
;;
;; 'I' First machine-dependent integer constant.
;; 'J' Second machine-dependent integer constant.
;; 'K' Third machine-dependent integer constant.
;; 'L' Fourth machine-dependent integer constant.
;; 'M' Fifth machine-dependent integer constant.
;; 'N' Sixth machine-dependent integer constant.
;; 'O' Seventh machine-dependent integer constant.
;; 'P' Eighth machine-dependent integer constant.
;;
;;     Other letters in the range `I' through `P' may be defined in a
;;     machine-dependent fashion to permit immediate integer operands with
;;     explicit integer values in specified ranges.  For example, on the 68000,
;;     `I' is defined to stand for the range of values 1 to 8.  This is the
;;     range permitted as a shift count in the shift instructions.
;;
;; `E' An immediate floating operand (expression code `const_double') is
;;     allowed, but only if the target floating point format is the same as
;;     that of the host machine (on which the compiler is running).
;;
;; `F' An immediate floating operand (expression code `const_double') is
;;     allowed.
;;
;; 'G' First machine-dependent const_double.
;; 'H' Second machine-dependent const_double.
;;
;; `s' An immediate integer operand whose value is not an explicit
;;     integer is allowed.
;;
;;     This might appear strange; if an insn allows a constant operand with a
;;     value not known at compile time, it certainly must allow any known
;;     value.  So why use `s' instead of `i'?  Sometimes it allows better code
;;     to be generated.
;;
;;     For example, on the 68000 in a fullword instruction it is possible to
;;     use an immediate operand; but if the immediate value is between -128 and
;;     127, better code results from loading the value into a register and
;;     using the register.  This is because the load into the register can be
;;     done with a `moveq' instruction.  We arrange for this to happen by
;;     defining the letter `K' to mean "any integer outside the range -128 to
;;     127", and then specifying `Ks' in the operand constraints.
;;
;; `g' Any register, memory or immediate integer operand is allowed,
;;     except for registers that are not general registers.
;;
;; `X' Any operand whatsoever is allowed, even if it does not satisfy
;;     `general_operand'.  This is normally used in the constraint of a
;;     `match_scratch' when certain alternatives will not actually require a
;;     scratch register.
;;
;; `0' Match operand 0.
;; `1' Match operand 1.
;; `2' Match operand 2.
;; `3' Match operand 3.
;; `4' Match operand 4.
;; `5' Match operand 5.
;; `6' Match operand 6.
;; `7' Match operand 7.
;; `8' Match operand 8.
;; `9' Match operand 9.
;;
;;     An operand that matches the specified operand number is allowed.  If a
;;     digit is used together with letters within the same alternative, the
;;     digit should come last.
;;
;;     This is called a "matching constraint" and what it really means is that
;;     the assembler has only a single operand that fills two roles considered
;;     separate in the RTL insn.  For example, an add insn has two input
;;     operands and one output operand in the RTL, but on most CISC machines an
;;     add instruction really has only two operands, one of them an
;;     input-output operand:
;;
;;          addl #35,r12
;;
;;     Matching constraints are used in these circumstances.  More precisely,
;;     the two operands that match must include one input-only operand and one
;;     output-only operand.  Moreover, the digit must be a smaller number than
;;     the number of the operand that uses it in the constraint.
;;
;;     For operands to match in a particular case usually means that they are
;;     identical-looking RTL expressions.  But in a few special cases specific
;;     kinds of dissimilarity are allowed.  For example, `*x' as an input
;;     operand will match `*x++' as an output operand.  For proper results in
;;     such cases, the output template should always use the output-operand's
;;     number when printing the operand.
;;
;; `p' An operand that is a valid memory address is allowed.  This is for
;;     "load address" and "push address" instructions.
;;
;;     `p' in the constraint must be accompanied by `address_operand' as the
;;     predicate in the `match_operand'.  This predicate interprets the mode
;;     specified in the `match_operand' as the mode of the memory reference for
;;     which the address would be valid.
;;
;; `Q` First non constant, non register machine-dependent insns
;; `R` Second non constant, non register machine-dependent insns
;; `S` Third non constant, non register machine-dependent insns
;; `T` Fourth non constant, non register machine-dependent insns
;; `U` Fifth non constant, non register machine-dependent insns
;;
;;     Letters in the range `Q' through `U' may be defined in a
;;     machine-dependent fashion to stand for arbitrary operand types.  The
;;     machine description macro `EXTRA_CONSTRAINT' is passed the operand as
;;     its first argument and the constraint letter as its second operand.
;;
;;     A typical use for this would be to distinguish certain types of memory
;;     references that affect other insn operands.
;;
;;     Do not define these constraint letters to accept register references
;;     (`reg'); the reload pass does not expect this and would not handle it
;;     properly.

;; Multiple Alternative Constraints
;; `?' Disparage slightly the alternative that the `?' appears in, as a
;;     choice when no alternative applies exactly.  The compiler regards this
;;     alternative as one unit more costly for each `?' that appears in it.
;;
;; `!' Disparage severely the alternative that the `!' appears in.  This
;;     alternative can still be used if it fits without reloading, but if
;;     reloading is needed, some other alternative will be used.

;; Constraint modifiers
;; `=' Means that this operand is write-only for this instruction: the
;;     previous value is discarded and replaced by output data.
;;
;; `+' Means that this operand is both read and written by the
;;     instruction.
;;
;;     When the compiler fixes up the operands to satisfy the constraints, it
;;     needs to know which operands are inputs to the instruction and which are
;;     outputs from it.  `=' identifies an output; `+' identifies an operand
;;     that is both input and output; all other operands are assumed to be
;;     input only.
;;
;; `&' Means (in a particular alternative) that this operand is written
;;     before the instruction is finished using the input operands.  Therefore,
;;     this operand may not lie in a register that is used as an input operand
;;     or as part of any memory address.
;;
;;     `&' applies only to the alternative in which it is written.  In
;;     constraints with multiple alternatives, sometimes one alternative
;;     requires `&' while others do not.
;;
;;     `&' does not obviate the need to write `='.
;;
;; `%' Declares the instruction to be commutative for this operand and the
;;     following operand.  This means that the compiler may interchange the two
;;     operands if that is the cheapest way to make all operands fit the
;;     constraints.  This is often used in patterns for addition instructions
;;     that really have only two operands: the result must go in one of the
;;     arguments.
;;
;; `#' Says that all following characters, up to the next comma, are to be
;;     ignored as a constraint.  They are significant only for choosing
;;     register preferences.
;;
;; `*' Says that the following character should be ignored when choosing
;;     register preferences.  `*' has no effect on the meaning of the
;;     constraint as a constraint, and no effect on reloading.


;; ::::::::::::::::::::
;; ::
;; :: Attributes
;; ::
;; ::::::::::::::::::::

;; The `define_attr' expression is used to define each attribute required by
;; the target machine.  It looks like:
;;
;; (define_attr NAME LIST-OF-VALUES DEFAULT)

;; NAME is a string specifying the name of the attribute being defined.

;; LIST-OF-VALUES is either a string that specifies a comma-separated list of
;; values that can be assigned to the attribute, or a null string to indicate
;; that the attribute takes numeric values.

;; DEFAULT is an attribute expression that gives the value of this attribute
;; for insns that match patterns whose definition does not include an explicit
;; value for this attribute.

;; For each defined attribute, a number of definitions are written to the
;; `insn-attr.h' file.  For cases where an explicit set of values is specified
;; for an attribute, the following are defined:

;; * A `#define' is written for the symbol `HAVE_ATTR_NAME'.
;;
;; * An enumeral class is defined for `attr_NAME' with elements of the
;;   form `UPPER-NAME_UPPER-VALUE' where the attribute name and value are first
;;   converted to upper case.
;;
;; * A function `get_attr_NAME' is defined that is passed an insn and
;;   returns the attribute value for that insn.

;; For example, if the following is present in the `md' file:
;;
;; (define_attr "type" "branch,fp,load,store,arith" ...)
;;
;; the following lines will be written to the file `insn-attr.h'.
;;
;; #define HAVE_ATTR_type
;; enum attr_type {TYPE_BRANCH, TYPE_FP, TYPE_LOAD, TYPE_STORE, TYPE_ARITH};
;; extern enum attr_type get_attr_type ();

;; If the attribute takes numeric values, no `enum' type will be defined and
;; the function to obtain the attribute's value will return `int'.

(define_attr "type"
  "load,branch,cbranch,bnzd,sfunc,pop,t_set,other"
  (const_string "other"))

(define_attr "load"
  "no,yes"
  (cond [(ne (symbol_ref
	      "memory_operand (SET_SRC (XVECEXP (PATTERN (insn), 0, 0)),
			       VOIDmode)")
	     (const_int 0))
	 (const_string "yes")]
	(const_string "no")))

(define_attr "is_sfunc" ""
  (if_then_else (eq_attr "type" "sfunc") (const_int 1) (const_int 0)))

;; Branches with 11-bit offset have a range of -1022 .. +1024, and a length
;; of two, i.e. the maximum offset from the end of the insn is 1022.
;; However, if the insn has a delay slot insn, shorten_branches counts
;; the length of a forward branch starting with the end of the delay slot
;; insn.  Since a delay slot insn can be up to four bytes (in the annulled
;; case), this reduces the safe forward range to 1018.
(define_attr "branch_o11_p" "no,yes"
  (cond [(leu (plus (minus (match_dup 0) (pc)) (const_int 1022))
	      (const_int 2040))
	 (const_string "yes")
	 (gtu (minus (match_dup 0) (pc)) (const_int 1022))
	 (const_string "no")
	 (ne (symbol_ref "NEXT_INSN (PREV_INSN (insn)) == insn") (const_int 0))
	 (const_string "yes")]
	(const_string "no")))

;; Bnzd has an 8-bit offset, giving a range of -126 .. + 128, and a length
;; of two, or four when the delay slot is not filled.
(define_attr "bnzd_o8_p" "no,yes"
  (cond [(leu (plus (minus (match_dup 0) (pc)) (const_int 126)) (const_int 250))
	 (const_string "yes")
	 (ne (minus (match_dup 0) (pc)) (const_int 126))
	 (const_string "no")
	 (ne (symbol_ref "NEXT_INSN (PREV_INSN (insn)) == insn") (const_int 0))
	 (const_string "yes")]
	(const_string "no")))

;; If that is out of range, we can fall back to sub / brt, giving
;; a length of 4.
(define_attr "bnzd_o11_p" "no,yes"
  (cond [(leu (plus (minus (match_dup 0) (pc)) (const_int 1022))
	      (const_int 2042))
	 (const_string "yes")]
	(const_string "no")))

;; ??? Using pc is not computed transitively, so we sprinkle in
;; '(ne (match_dup 0) (match_dup 0))' clauses for genattrtab to get the point.
;; FIXME: a long cbranch gets even longer if it has a t_set delay slot insn.
(define_attr "length" ""
  (cond [(eq_attr "type" "branch,cbranch")
	 (cond [(eq_attr "branch_o11_p" "yes") (const_int 2)
		(ne (match_dup 0) (match_dup 0)) (const_int 4)] (const_int 4))
	 (eq_attr "type" "bnzd")
	 (cond [(eq_attr "bnzd_o8_p" "yes") (const_int 2)
		(eq_attr "bnzd_o11_p" "yes") (const_int 4)
		(ne (match_dup 0) (match_dup 0)) (const_int 6)] (const_int 6))
	 (eq_attr "type" "sfunc")
	 (const_int 4)]
	(const_int 2)))

;; Does this insn load a register (as operand 0), which might be r6 / r7,
;; so that if it is register 6 or 7, Z0 / Z1 is set accordingly?
(define_attr "loop_count_set" "no,yes" (const_string "no"))


;; ::::::::::::::::::::
;; ::
;; :: Function Units
;; ::
;; ::::::::::::::::::::

;; On most RISC machines, there are instructions whose results are not
;; available for a specific number of cycles.  Common cases are instructions
;; that load data from memory.  On many machines, a pipeline stall will result
;; if the data is referenced too soon after the load instruction.

;; In addition, many newer microprocessors have multiple function units,
;; usually one for integer and one for floating point, and often will incur
;; pipeline stalls when a result that is needed is not yet ready.

;; The descriptions in this section allow the specification of how much time
;; must elapse between the execution of an instruction and the time when its
;; result is used.  It also allows specification of when the execution of an
;; instruction will delay execution of similar instructions due to function
;; unit conflicts.

;; For the purposes of the specifications in this section, a machine is divided
;; into "function units", each of which execute a specific class of
;; instructions in first-in-first-out order.  Function units that accept one
;; instruction each cycle and allow a result to be used in the succeeding
;; instruction (usually via forwarding) need not be specified.  Classic RISC
;; microprocessors will normally have a single function unit, which we can call
;; `memory'.  The newer "superscalar" processors will often have function units
;; for floating point operations, usually at least a floating point adder and
;; multiplier.

;; Each usage of a function units by a class of insns is specified with a
;; `define_function_unit' expression, which looks like this:

;; (define_function_unit NAME MULTIPLICITY SIMULTANEITY TEST READY-DELAY
;;   ISSUE-DELAY [CONFLICT-LIST])

;; NAME is a string giving the name of the function unit.

;; MULTIPLICITY is an integer specifying the number of identical units in the
;; processor.  If more than one unit is specified, they will be scheduled
;; independently.  Only truly independent units should be counted; a pipelined
;; unit should be specified as a single unit.  (The only common example of a
;; machine that has multiple function units for a single instruction class that
;; are truly independent and not pipelined are the two multiply and two
;; increment units of the CDC 6600.)

;; SIMULTANEITY specifies the maximum number of insns that can be executing in
;; each instance of the function unit simultaneously or zero if the unit is
;; pipelined and has no limit.

;; All `define_function_unit' definitions referring to function unit NAME must
;; have the same name and values for MULTIPLICITY and SIMULTANEITY.

;; TEST is an attribute test that selects the insns we are describing in this
;; definition.  Note that an insn may use more than one function unit and a
;; function unit may be specified in more than one `define_function_unit'.

;; READY-DELAY is an integer that specifies the number of cycles after which
;; the result of the instruction can be used without introducing any stalls.

;; ISSUE-DELAY is an integer that specifies the number of cycles after the
;; instruction matching the TEST expression begins using this unit until a
;; subsequent instruction can begin.  A cost of N indicates an N-1 cycle delay.
;; A subsequent instruction may also be delayed if an earlier instruction has a
;; longer READY-DELAY value.  This blocking effect is computed using the
;; SIMULTANEITY, READY-DELAY, ISSUE-DELAY, and CONFLICT-LIST terms.  For a
;; normal non-pipelined function unit, SIMULTANEITY is one, the unit is taken
;; to block for the READY-DELAY cycles of the executing insn, and smaller
;; values of ISSUE-DELAY are ignored.

;; CONFLICT-LIST is an optional list giving detailed conflict costs for this
;; unit.  If specified, it is a list of condition test expressions to be
;; applied to insns chosen to execute in NAME following the particular insn
;; matching TEST that is already executing in NAME.  For each insn in the list,
;; ISSUE-DELAY specifies the conflict cost; for insns not in the list, the cost
;; is zero.  If not specified, CONFLICT-LIST defaults to all instructions that
;; use the function unit.

;; Typical uses of this vector are where a floating point function unit can
;; pipeline either single- or double-precision operations, but not both, or
;; where a memory unit can pipeline loads, but not stores, etc.

;; As an example, consider a classic RISC machine where the result of a load
;; instruction is not available for two cycles (a single "delay" instruction is
;; required) and where only one load instruction can be executed
;; simultaneously.  This would be specified as:

;; (define_function_unit "memory" 1 1 (eq_attr "type" "load") 2 0)

;; For the case of a floating point function unit that can pipeline
;; either single or double precision, but not both, the following could be
;; specified:
;;
;; (define_function_unit "fp" 1 0
;;   (eq_attr "type" "sp_fp") 4 4
;;   [(eq_attr "type" "dp_fp")])
;;
;; (define_function_unit "fp" 1 0
;;   (eq_attr "type" "dp_fp") 4 4
;;   [(eq_attr "type" "sp_fp")])

;; Note: The scheduler attempts to avoid function unit conflicts and uses all
;; the specifications in the `define_function_unit' expression.  It has
;; recently come to our attention that these specifications may not allow
;; modeling of some of the newer "superscalar" processors that have insns using
;; multiple pipelined units.  These insns will cause a potential conflict for
;; the second unit used during their execution and there is no way of
;; representing that conflict.  We welcome any examples of how function unit
;; conflicts work in such processors and suggestions for their representation.

(define_function_unit "memory" 1 0 (eq_attr "type" "load,pop") 2 1)


;; ::::::::::::::::::::
;; ::
;; :: Delay Slots
;; ::
;; ::::::::::::::::::::

;; The insn attribute mechanism can be used to specify the requirements for
;; delay slots, if any, on a target machine.  An instruction is said to require
;; a "delay slot" if some instructions that are physically after the
;; instruction are executed as if they were located before it.  Classic
;; examples are branch and call instructions, which often execute the following
;; instruction before the branch or call is performed.

;; On some machines, conditional branch instructions can optionally "annul"
;; instructions in the delay slot.  This means that the instruction will not be
;; executed for certain branch outcomes.  Both instructions that annul if the
;; branch is true and instructions that annul if the branch is false are
;; supported.

;; Delay slot scheduling differs from instruction scheduling in that
;; determining whether an instruction needs a delay slot is dependent only
;; on the type of instruction being generated, not on data flow between the
;; instructions.  See the next section for a discussion of data-dependent
;; instruction scheduling.

;; The requirement of an insn needing one or more delay slots is indicated via
;; the `define_delay' expression.  It has the following form:
;;
;; (define_delay TEST
;;   [DELAY-1 ANNUL-TRUE-1 ANNUL-FALSE-1
;;    DELAY-2 ANNUL-TRUE-2 ANNUL-FALSE-2
;;    ...])

;; TEST is an attribute test that indicates whether this `define_delay' applies
;; to a particular insn.  If so, the number of required delay slots is
;; determined by the length of the vector specified as the second argument.  An
;; insn placed in delay slot N must satisfy attribute test DELAY-N.
;; ANNUL-TRUE-N is an attribute test that specifies which insns may be annulled
;; if the branch is true.  Similarly, ANNUL-FALSE-N specifies which insns in
;; the delay slot may be annulled if the branch is false.  If annulling is not
;; supported for that delay slot, `(nil)' should be coded.

;; For example, in the common case where branch and call insns require a single
;; delay slot, which may contain any insn other than a branch or call, the
;; following would be placed in the `md' file:

;; (define_delay (eq_attr "type" "branch,call")
;;		 [(eq_attr "type" "!branch,call") (nil) (nil)])

;; Multiple `define_delay' expressions may be specified.  In this case, each
;; such expression specifies different delay slot requirements and there must
;; be no insn for which tests in two `define_delay' expressions are both true.

;; For example, if we have a machine that requires one delay slot for branches
;; but two for calls, no delay slot can contain a branch or call insn, and any
;; valid insn in the delay slot for the branch can be annulled if the branch is
;; true, we might represent this as follows:

;; (define_delay (eq_attr "type" "branch")
;;   [(eq_attr "type" "!branch,call")
;;    (eq_attr "type" "!branch,call")
;;    (nil)])
;;
;; (define_delay (eq_attr "type" "call")
;;   [(eq_attr "type" "!branch,call") (nil) (nil)
;;    (eq_attr "type" "!branch,call") (nil) (nil)])

;; Note - it is the backend's responsibility to fill any unfilled delay slots
;; at assembler generation time.  This is usually done by adding a special print
;; operand to the delayed insrtuction, and then in the PRINT_OPERAND function
;; calling dbr_sequence_length() to determine how many delay slots were filled.
;; For example:
;;
;; --------------generic.md-----------------
;; (define_insn "call"
;;  [(call (match_operand 0 "memory_operand" "m")
;;         (match_operand 1 "" ""))]
;;   ""
;;   "call_delayed %0,%1,%2%#"
;;  [(set_attr "length" "4")
;;   (set_attr "type" "call")])
;;
;; -------------generic.h-------------------
;; #define PRINT_OPERAND_PUNCT_VALID_P(CODE) (CODE == '#')
;;
;;  ------------generic.c------------------
;; void
;; generic_print_operand (file, x, code)
;;     FILE * file;
;;     rtx    x;
;;     int    code;
;; {
;;   switch (code)
;;   {
;;   case '#':
;;     if (dbr_sequence_length () == 0)
;;       fputs ("\n\tnop", file);
;;     return;

(define_attr "optional_delay_slot" "yes,no"
  (cond [(eq_attr "type" "branch") (const_string "yes")] (const_string "no")))

(define_attr "cbranch_delay_slot" "yes,no"
  (cond [(eq_attr "type" "cbranch") (const_string "yes")] (const_string "no")))

(define_attr "return_delay_slot" "yes,no" (const_string "no"))

(define_attr "needs_delay_slot" "yes,no" (const_string "no"))

(define_attr "wants_delay_slot" "yes,no"
  (cond [(eq_attr "optional_delay_slot" "yes") (const_string "yes")
	 (eq_attr "return_delay_slot" "yes") (const_string "yes")
	 (eq_attr "needs_delay_slot" "yes") (const_string "yes")]
	(const_string "no")))

(define_attr "in_delay_slot" "yes,no"
  (cond [(eq_attr "wants_delay_slot" "yes") (const_string "no")
	 (eq_attr "type" "cbranch") (const_string "no")
	 (eq_attr "length" "2") (const_string "yes")]
        (const_string "no")))

;; delay slot scheduling of t-setting insns from the fall-through chain can
;; make some very nice optimizations, but eager delay slot scheduling is
;; less interesting.  Moreover, cbranches cannot simply put t-setting insns
;; in front if the delay slot falls victim to an out-of-range offset, so try
;; to avoid eager delay slot scheduling for cbranches with t-setting insns.
(define_attr "in_cbranch_delay_slot" "yes,no"
  (cond [(eq_attr "in_delay_slot" "no") (const_string "no")
	 (eq_attr "type" "!t_set") (const_string "yes")
	 (ne (symbol_ref "GET_CODE (prev_nonnote_insn (insn)) == CODE_LABEL")
	     (const_int 0))
	 (const_string "no")]
	(const_string "yes")))

(define_attr "in_return_delay_slot" "yes,no"
  (cond [(eq_attr "in_delay_slot" "no") (const_string "no")
	 (eq_attr "type" "!pop") (const_string "yes")
	 (ne (symbol_ref "dead_or_set_regno_p (insn, LINK_REG)") (const_int 0))
	 (const_string "no")]
        (const_string "yes")))

(define_delay
  (eq_attr "optional_delay_slot" "yes")
  [(eq_attr "in_delay_slot" "yes") (const_int 1) (nil)])

(define_delay
  (eq_attr "cbranch_delay_slot" "yes")
  [(eq_attr "in_cbranch_delay_slot" "yes") (const_int 1) (nil)])

(define_delay
  (eq_attr "return_delay_slot" "yes")
  [(eq_attr "in_return_delay_slot" "yes") (const_int 1) (nil)])

(define_delay
  (eq_attr "needs_delay_slot" "yes")
  [(eq_attr "in_delay_slot" "yes") (nil) (nil)])

;; ::::::::::::::::::::
;; ::
;; :: Notes on Patterns
;; ::
;; ::::::::::::::::::::

;; If you need to construct a sequence of assembler instructions in order
;; to implement a pattern be sure to escape any backslashes and double quotes
;; that you use, eg:
;;
;; (define_insn "an example"
;;   [(some rtl)]
;;   ""
;;   "*
;;    { static char buffer [100];
;;      sprintf (buffer, \"insn \\t %d\", REGNO (operands[1]));
;;      return buffer;
;;    }"
;; )
;;
;; Also if there is more than one instruction, they can be seperated by \\;
;; which is a space saving synonym for \\n\\t:
;;
;; (define_insn "another example"
;;   [(some rtl)]
;;   ""
;;   "*
;;    { static char buffer [100];
;;      sprintf (buffer, \"insn1 \\t %d\\;insn2 \\t %%1\",
;;        REGNO (operands[1]));
;;      return buffer;
;;    }"
;; )
;;


;; ::::::::::::::::::::
;; ::
;; :: Moves
;; ::
;; ::::::::::::::::::::

;; Wrap moves in define_expand to prevent memory->memory moves from being
;; generated at the RTL level, which generates better code for most machines
;; which can't do mem->mem moves.

;; If operand 0 is a `subreg' with mode M of a register whose own mode is wider
;; than M, the effect of this instruction is to store the specified value in
;; the part of the register that corresponds to mode M.  The effect on the rest
;; of the register is undefined.

;; This class of patterns is special in several ways.  First of all, each of
;; these names *must* be defined, because there is no other way to copy a datum
;; from one place to another.

;; Second, these patterns are not used solely in the RTL generation pass.  Even
;; the reload pass can generate move insns to copy values from stack slots into
;; temporary registers.  When it does so, one of the operands is a hard
;; register and the other is an operand that can need to be reloaded into a
;; register.

;; Therefore, when given such a pair of operands, the pattern must
;; generate RTL which needs no reloading and needs no temporary
;; registers--no registers other than the operands.  For example, if
;; you support the pattern with a `define_expand', then in such a
;; case the `define_expand' mustn't call `force_reg' or any other such
;; function which might generate new pseudo registers.

;; This requirement exists even for subword modes on a RISC machine
;; where fetching those modes from memory normally requires several
;; insns and some temporary registers.  Look in `spur.md' to see how
;; the requirement can be satisfied.

;; During reload a memory reference with an invalid address may be passed as an
;; operand.  Such an address will be replaced with a valid address later in the
;; reload pass.  In this case, nothing may be done with the address except to
;; use it as it stands.  If it is copied, it will not be replaced with a valid
;; address.  No attempt should be made to make such an address into a valid
;; address and no routine (such as `change_address') that will do so may be
;; called.  Note that `general_operand' will fail when applied to such an
;; address.
;;
;; The global variable `reload_in_progress' (which must be explicitly declared
;; if required) can be used to determine whether such special handling is
;; required.
;;
;; The variety of operands that have reloads depends on the rest of
;; the machine description, but typically on a RISC machine these can
;; only be pseudo registers that did not get hard registers, while on
;; other machines explicit memory references will get optional
;; reloads.
;;
;; If a scratch register is required to move an object to or from memory, it
;; can be allocated using `gen_reg_rtx' prior to reload.  But this is
;; impossible during and after reload.  If there are cases needing scratch
;; registers after reload, you must define `SECONDARY_INPUT_RELOAD_CLASS' and
;; perhaps also `SECONDARY_OUTPUT_RELOAD_CLASS' to detect them, and provide
;; patterns `reload_inM' or `reload_outM' to handle them.

;; The constraints on a `moveM' must permit moving any hard register to any
;; other hard register provided that `HARD_REGNO_MODE_OK' permits mode M in
;; both registers and `REGISTER_MOVE_COST' applied to their classes returns a
;; value of 2.

;; It is obligatory to support floating point `moveM' instructions
;; into and out of any registers that can hold fixed point values,
;; because unions and structures (which have modes `SImode' or
;; `DImode') can be in those registers and they may have floating
;; point members.

;; There may also be a need to support fixed point `moveM' instructions in and
;; out of floating point registers.  Unfortunately, I have forgotten why this
;; was so, and I don't know whether it is still true.  If `HARD_REGNO_MODE_OK'
;; rejects fixed point values in floating point registers, then the constraints
;; of the fixed point `moveM' instructions must be designed to avoid ever
;; trying to reload into a floating point register.

(define_expand "movqi"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
	(match_operand:QI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (QImode, operands[1]);
  emit_insn (gen_movqi_internal (operands[0], operands[1]));
  DONE;
}")

(define_insn "movqi_internal"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=r,d,r,d,Q,m,t,t,?r")
	(match_operand:QI 1 "general_operand"      "i,rQ,r,rm,d,d,J,N,?t"))
   (clobber (match_scratch:PQI 2 "=X,X,X,X,X,X,X,X,X"))]
  "risc16_move_check (operands)"
  "@
	%L1 %0,(%1 & 0xff)
	%L1 %0,%1
	%L1 %0,%1
	%L1 %0,%1
	ldb %0,%1
	ldb %0,%1
	cmp gt,r15,r15
	cmp eq,r15,r15
	sbc %0,%0% com %0% sra %0"
  [(set (attr "type") (cond [(eq_attr "load" "yes")
			    (const_string "load")] (const_string "other")))
   (set_attr "length" "2,2,2,4,2,4,2,2,6")
   (set_attr_alternative "loop_count_set"
     [(const_string "no")
      (cond [(eq_attr "load" "yes") (const_string "no")] (const_string "yes"))
      (const_string "yes")
      (cond [(eq_attr "load" "yes") (const_string "no")] (const_string "yes"))
      (const_string "no")
      (const_string "no")
      (const_string "no")
      (const_string "no")
      (const_string "yes")])])

(define_expand "movhi"
  [(set (match_operand:HI 0 "general_operand" "")
	(match_operand:HI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (HImode, operands[1]);
  /* When accessing a PSImode value inside a DImode structure, unfortunately
     store_bit_field will chop the store into two halves.  The access of
     the high word would, if unmodified, would be interpreted as an access
     of the low word at reload time.  */
  if (GET_CODE (operands[0]) == SUBREG)
    {
      rtx sub = SUBREG_REG (operands[0]);

      if (GET_MODE (sub) == PSImode && SUBREG_WORD (operands[0]) == 0)
	operands[0] = risc16_split_operand (sub, 2, 0);
    }
  /* If we have (mem:HI (addressof:PSI (reg:PSI ..))) as destination, force
     the address into a pseudo.  */
  if (GET_CODE (operands[0]) == MEM
      && GET_CODE (XEXP (operands[0], 0)) == ADDRESSOF
      && GET_MODE (XEXP (XEXP (operands[0], 0), 0)) == PSImode)
    operands[0]
      = change_address (operands[0], VOIDmode,
			force_reg (PSImode, XEXP (operands[0], 0)));
  if (GET_CODE (operands[1]) == SUBREG)
    {
      rtx sub = SUBREG_REG (operands[1]);

      if (sub == virtual_stack_vars_rtx)
	{
	  rtx reg = gen_reg_rtx (PSImode);

	  if (GET_MODE (sub) != PSImode)
	    abort ();
	  emit_insn (gen_addpsi3 (reg, sub, const0_rtx));
	  operands[1] = gen_rtx_SUBREG (HImode, reg, SUBREG_WORD (operands[1]));
	}
      if (GET_MODE (sub) == PSImode && SUBREG_WORD (operands[1]) == 0)
	operands[1] = risc16_split_operand (sub, 2, 0);
    }
  emit_insn (gen_movhi_i (operands[0], operands[1]));
  DONE;
}")

(define_expand "reload_load_address"
  [(set (match_operand 0 "" "") (match_operand 1 "" ""))]
  ""
  "
{
  if (GET_MODE (operands[0]) == HImode)
    {
      emit_insn (gen_movhi_i (operands[0], operands[1]));
      DONE;
    }
}")

;; ??? reload won't replace reg_equiv_constant unless a pseudo is
;; an operand.  So we need an extra pattern for moves from extension to
;; data registers.  Testcase: execute/stdio-opt-2.c -O2.
(define_insn "*movhi_e2d"
  [(set (match_operand:HI 0 "register_operand" "=d,d,d,d")
	(zero_extract:HI (match_operand:PSI 1 "nonimmediate_operand" "a,Q,i,m")
			 (const_int 6) (const_int 16)))
   (clobber (match_scratch:PQI 2 "=X,X,X,X"))]
  ""
  "@
	ld %0,%E1
	ldw %0,%1
	ld %0,#%%hi(%1)
	ldw %0,%1"
  [(set_attr "type"         "other,load,other,load")
   (set_attr "length"         "2,2,4,4")])

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "movhi_i"
  [(set (match_operand:HI 0 "risc16_movhidst_operand"
	    "=r,r,Q,m,t,t,t,r,S,d")
	(match_operand:HI 1 "risc16_movhisrc_operand"
	 "rLQ,rim,r,r,r,J,N,??t,d,S"))
   (clobber (match_scratch:PQI 2 "=X,X,X,X,X,X,X,X,X,X"))]
  "risc16_move_check (operands)"
  "@
	%L1 %0,%1
	%L1 %0,%R1
	ldw %0,%1
	ldw %0,%1
	sra %1
	cmp gt,r15,r15
	cmp eq,r15,r15
	sbc %0,%0% com %0% sra %0
	%L0 %0,%1
	%L1 %0,%1"
  [(set_attr_alternative "type"
     [(cond [(eq_attr "load" "yes") (const_string "load")]
	    (const_string "other"))
      (cond [(eq_attr "load" "yes") (const_string "load")]
	    (const_string "other"))
      (const_string "other")
      (const_string "other")
      (const_string "t_set")
      (const_string "t_set")
      (const_string "t_set")
      (const_string "other")
      (const_string "other")
      (const_string "other")])
   (set_attr "length"         "2,4,2,4,2,2,2,6,2,2")
   (set_attr_alternative "loop_count_set"
     [(cond [(eq_attr "load" "yes") (const_string "no")] (const_string "yes"))
      (cond [(eq_attr "load" "yes") (const_string "no")] (const_string "yes"))
      (const_string "no")
      (const_string "no")
      (const_string "no")
      (const_string "no")
      (const_string "no")
      (const_string "yes")
      (const_string "no")
      (const_string "no")])])

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "set_loopcount"
  [(set (match_operand:HI 0 "register_operand" "=j,j,k,k,j,k")
	(match_operand:HI 1 "general_operand" "rL,i,rL,i,t,t"))
   (set (match_operand:PQI 2 "register_operand" "=v,v,w,w,v,w")
	(eq:PQI (match_dup 1) (const_int 0)))]
  ""
  "@
	ld %0,%1
	ld %0,%1
	ld %0,%1
	ld %0,%1
	sbc %0,%0% com %0% sra %0
	sbc %0,%0% com %0% sra %0"
  [(set_attr "length" "2,4,2,4,6,6")])

;; post-reload splitters can leave meandering moves - straighten them up.
(define_peephole2
  [(parallel [(set (match_operand:HI 0 "register_operand" "")
		   (match_operand:HI 1 "general_operand" ""))
	      (clobber (match_scratch:PQI 2 ""))])
   (parallel [(set (match_operand:HI 3 "nonimmediate_operand" "") (match_dup 0))
	      (clobber (match_scratch:PQI 4 ""))])]
  "(arith_reg_operand (operands[1], HImode)
    || arith_reg_operand(operands[3], HImode))
   && peep2_reg_dead_p (2, operands[0])"
  [(parallel [(set (match_dup 3) (match_dup 1))
	      (clobber (match_dup 4))])]
  "")

(define_expand "movsi"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
	(match_operand:SI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (SImode, operands[1]);
  emit_insn (gen_movsi_internal (operands[0], operands[1]));
  DONE;
}")

;; Note - it is best to only have one movsi pattern and to handle
;; all the various contingencies by the use of alternatives.  This
;; allows reload the greatest amount of flexability (since reload will
;; only choose amoungst alternatives for a selected insn, it will not
;; replace the insn with another one).

;; Naturally, SImode, PSImode and DImode moves to data regs would require
;; up to two clobbers to clobber both Z0 and Z1 when r6 and r7 are used.
;; However, reload patterns can only handle a single clobber, and there
;; is no mode to encompass two PQImode registers, so we just pretend these
;; wide moves don't clobber Z0 / Z1 - that also makes the patterns much
;; simpler.  We split all the alternatives that set data registers (and
;; then some) after reload, and add the clobbers there.  That works because
;; Z0 / Z1 become only live in machine_dependent_reorg.

(define_insn_and_split "movsi_internal"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=r,m")
	(match_operand:SI 1 "general_operand"       "rim,r"))]
  "!memory_operand (operands[0], SImode) || !memory_operand (operands[1], SImode)"
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_move (operands[0], operands[1]); DONE;")

(define_expand "movpsi"
  [(set (match_operand:PSI 0 "nonimmediate_operand" "")
	(match_operand:PSI 1 "general_operand" ""))]
  ""
  "
{
  rtx insn;

  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (PSImode, operands[1]);
  if (GET_CODE (operands[1]) == REG
      && ((REGNO (operands[1]) >= FIRST_VIRTUAL_REGISTER
	   && REGNO (operands[1]) <= LAST_VIRTUAL_REGISTER)
	  || (operands[1] == frame_pointer_rtx
	      && ! reload_completed && ! reload_in_progress)
	  || operands[1] == arg_pointer_rtx))
    {
      rtx reg = gen_reg_rtx (PSImode);

      emit_insn (gen_addpsi3 (reg, operands[1], const0_rtx));
      operands[1] = reg;
    }
  if (GET_CODE (operands[1]) == SUBREG
      && GET_MODE (XEXP (operands[1], 0)) != PSImode)
    {
      rtx inner_reg = SUBREG_REG (operands[1]);

      if (GET_MODE (inner_reg) == SImode)
	{
	  if (SUBREG_WORD (operands[1]) != 0)
	    abort ();
	}
      else
	inner_reg
	  = gen_rtx_SUBREG (SImode, inner_reg, SUBREG_WORD (operands[1]));
      emit_insn (gen_truncsipsi2 (operands[0], inner_reg));
      DONE;
    }
  if (GET_CODE (operands[1]) == LABEL_REF
      && GET_CODE (XEXP (operands[1], 0)) == NOTE)
    operands[1] = const0_rtx;
  insn = emit_insn (gen_movpsi_i (operands[0], operands[1]));
  if (reload_in_progress
      && GET_CODE (operands[1]) == LABEL_REF)
    {
      REG_NOTES (insn)
	= gen_rtx_EXPR_LIST (REG_LABEL, XEXP (operands[1], 0),
			     REG_NOTES (insn));
      LABEL_NUSES (XEXP (operands[1], 0))++;
    }
  DONE;
}")

;; The a / ma alternative is basically a / m, but allows reload inheritance.
;; Likewise for d / dio.
(define_insn_and_split "movpsi_i"
  [(set (match_operand:PSI 0 "risc16_movdst_operand" "=a,a,d,a,a,Q,a,m,o,e")
	(match_operand:PSI 1 "risc16_movsrc_operand" "a,aQ,a,d,i,a,ma,a,e,eio"))]
  "!memory_operand (operands[0], PSImode) || !memory_operand (operands[1], PSImode)"
  "@
	ld %0,%1
	%L1 %0,%1
	#
	ld r%N0,%R1\;ld %E0,%S1
	ld %0,%1
	ldw %0,%1
	%L1 %0,%1
	ldw %0,%1
	#
	#"
  "reload_completed
   && ((GET_CODE (operands[0]) == REG
	&& REGNO (operands[0]) <= LAST_DATA_REG)
       || (GET_CODE (operands[1]) == REG
	   && REGNO (operands[1]) <= LAST_DATA_REG))"
  [(pc)]
  "risc16_split_move (operands[0], operands[1]); DONE;"
  [(set_attr "type" "other,load,other,other,other,other,load,other,other,load")
   (set_attr "length" "2,2,4,4,4,2,4,4,8,8")])

(define_expand "movdi"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
	(match_operand:DI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (DImode, operands[1]);
  emit_insn (gen_movdi_internal (operands[0], operands[1]));
  DONE;
}")

(define_insn_and_split "movdi_internal"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=r,m")
	(match_operand:DI 1 "general_operand"       "rim,r"))]
  "!memory_operand (operands[0], DImode) || !memory_operand (operands[1], DImode)"
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_move (operands[0], operands[1]); DONE;")

;; Floating Point Moves
;;
;; Note - Patterns for SF mode moves are compulsory, but
;; patterns for DF are optional, as GCC can synthesise them.

(define_expand "movsf"
  [(set (match_operand:SF 0 "general_operand" "")
	(match_operand:SF 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (SFmode, operands[1]);
  emit_insn (gen_movsf_internal (operands[0], operands[1]));
  DONE;
}")

(define_insn_and_split "movsf_internal"
  [(set (match_operand:SF 0 "nonimmediate_operand" "=r,m")
	(match_operand:SF 1 "general_operand" "rim,r"))]
  "!memory_operand (operands[0], SFmode) || !memory_operand (operands[1], SFmode)"
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_move (operands[0], operands[1]); DONE;")

;; (define_expand "movdf"
;;   [(set (match_operand:DF 0 "general_operand" "")
;; 	(match_operand:DF 1 "general_operand" ""))]
;;   ""
;;   "
;; {
;;  if (!reload_in_progress
;;      && !reload_completed
;;      && GET_CODE (operands[0]) == MEM
;;      && GET_CODE (operands[1]) == MEM)
;;     operands[1] = copy_to_mode_reg (DFmode, operands[1]);
;; }")
;; 
;; (define_insn "*movdf_internal"
;;   [(set (match_operand:DF 0 "general_operand" "=g")
;; 	(match_operand:DF 1 "general_operand" "g"))]
;;   "!memory_operand (operands[0], DFmode) || !memory_operand (operands[1], DFmode)"
;;   "movdf %0,%1"
;;   [(set_attr "length" "4")])


;; ::::::::::::::::::::
;; ::
;; :: Conversions
;; ::
;; ::::::::::::::::::::

;; Signed conversions from a smaller integer to a larger integer
;;
;; These operations are optional.  If they are not
;; present GCC will synthesise them for itself

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(sign_extend:HI (match_operand:QI 1 "general_operand" "0")))
   (clobber (match_scratch:PQI 2 "=X"))]
  ""
  "ext %0"
  [(set_attr "loop_count_set" "yes")])

;;(define_insn "extendqisi2"
;;  [(set (match_operand:SI 0 "register_operand" "=r")
;;	(sign_extend:SI (match_operand:QI 1 "general_operand" "g")))]
;;  ""
;;  "extendqisi2 %0,%1"
;;  [(set_attr "length" "4")])
;;
;;(define_insn "extendqidi2"
;;  [(set (match_operand:DI 0 "register_operand" "=r")
;;	(sign_extend:DI (match_operand:QI 1 "general_operand" "g")))]
;;  ""
;;  "extendqihi2 %0,%1"
;;  [(set_attr "length" "4")])

(define_insn_and_split "extendhipsi2"
  [(set (match_operand:PSI 0 "register_operand" "=a")
	(sign_extend:PSI (match_operand:HI 1 "general_operand" "0")))
   (clobber (match_scratch:HI 2 "=d"))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx high = risc16_split_operand (operands[0], 2, 0);
  rtx low = gen_lowpart (HImode, operands[0]);

  emit_insn (gen_cmpgehi_t (low, const0_rtx));
  emit_hi_insn (gen_ldnott (operands[2]));
  risc16_split_move (high, operands[2]);
  DONE;
}")

(define_insn_and_split "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=d,?r,d,?r")
	(sign_extend:SI (match_operand:HI 1 "general_operand" "0,0,?g,?g")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx high = gen_highpart (HImode, operands[0]);
  rtx low = gen_lowpart (HImode, operands[0]);

  if (! operands_match_p (operands[0], operands[1]))
    risc16_split_move (low, operands[1]);
  emit_insn (gen_cmpgehi_t (low, const0_rtx));
  emit_hi_insn (gen_ldnott (high));
  DONE;
}")

(define_insn_and_split "extendhidi2"
  [(set (match_operand:DI 0 "register_operand" "=d,?r,d,?r")
	(sign_extend:DI (match_operand:HI 1 "general_operand" "0,0,??g,??g")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx high0 = risc16_split_operand (operands[0], 4, 0);
  rtx high1 = risc16_split_operand (operands[0], 4, 1);
  rtx med = risc16_split_operand (operands[0], 4, 2);
  rtx low = risc16_split_operand (operands[0], 4, 3);

  if (! operands_match_p (operands[0], operands[1]))
    risc16_split_move (low, operands[1]);
  emit_insn (gen_cmpgehi_t (low, const0_rtx));
  emit_hi_insn (gen_ldnott (med));
  emit_hi_insn (gen_ldnott (high1));
  emit_hi_insn (gen_ldnott (high0));
  DONE;
}")

(define_insn_and_split "extendsidi2"
  [(set (match_operand:DI 0 "register_operand" "=d,?r,d,?r")
	(sign_extend:DI (match_operand:SI 1 "general_operand" "0,0,??g,??g")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx high0 = risc16_split_operand (operands[0], 4, 0);
  rtx high1 = risc16_split_operand (operands[0], 4, 1);
  rtx med = risc16_split_operand (operands[0], 4, 2);
  rtx low = risc16_split_operand (operands[0], 4, 3);

  if (! operands_match_p (operands[0], operands[1]))
    {
      if (REGNO (low) == REGNO (operands[1]))
	{
	  risc16_split_move (med, risc16_split_operand (operands[1], 2, 0));
	  risc16_split_move (low, risc16_split_operand (operands[1], 2, 1));
	}
      else
	{
	  risc16_split_move (low, risc16_split_operand (operands[1], 2, 1));
	  risc16_split_move (med, risc16_split_operand (operands[1], 2, 0));
	}
    }
  emit_insn (gen_cmpgehi_t (med, const0_rtx));
  emit_hi_insn (gen_ldnott (high1));
  emit_hi_insn (gen_ldnott (high0));
  DONE;
}")

;;;; Unsigned conversions from a smaller integer to a larger integer
;; Don't use 'z' for reloading, because if that hard register happens 
;; to be used, we fail.  ??? Why can't reload disparage alternatives
;; where most / all registers are not available?

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=!z,r,&r,d")
	(zero_extend:HI (match_operand:QI 1 "general_operand" "0,0,r,m")))
   (clobber (match_scratch:PQI 2 "=X,X,X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "@
	and %0,#255
	and %0,#255
	ld %0,#255% and %0,%1
	ldb %0,%1"
  [(set_attr "type" "t_set,t_set,t_set,load")
   (set_attr "length" "2,4,4,4")
   (set_attr "loop_count_set" "no,yes,yes,no")
   (set_attr "type" "t_set")])

(define_insn_and_split "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,d")
	(zero_extend:SI (match_operand:QI 1 "general_operand" "r,ro")))]
  ""
  "#"
  ""
  [(pc)]
  "
{
  emit_insn (gen_zero_extendqihi2 (operand_subword (operands[0], 1, 0, SImode),
				   operands[1]));
  emit_move_insn (operand_subword (operands[0], 0, 0, SImode), const0_rtx);
  DONE;
}")

;;(define_insn "zero_extendqidi2"
;;  [(set (match_operand:DI 0 "register_operand" "=r")
;;	(zero_extend:DI (match_operand:QI 1 "general_operand" "g")))]
;;  ""
;;  "zero_extendqihi2 %0,%1"
;;  [(set_attr "length" "4")])

(define_insn_and_split "zero_extendhipsi2"
  [(set (match_operand:PSI 0 "register_operand" "=a")
	(zero_extend:PSI (match_operand:HI 1 "register_operand" "0")))
   (clobber (match_scratch:HI 2 "=d"))]
  ""
  "ld %2,#0% ld %E0,%2"
  "reload_completed"
  [(pc)]
  "
{
  emit_insn (gen_movhi_i (operands[2], const0_rtx));
  emit_insn (gen_movhi_i (risc16_split_operand(operands[0], 2, 0),
			  operands[2]));
  DONE;
}")

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=x,r,x,r,x,r")
	(zero_extend:SI (match_operand:HI 1 "general_operand" "0,0,?r,?r,??g,??g")))
   (clobber (match_scratch:PQI 2 "=X,c,X,c,X,c"))
   (clobber (match_scratch:PQI 3 "=X,c,X,c,X,c"))]
  ""
  "@
	ld %S0,#0
	ld %S0,#0
	%L1 %R0,%1% ld %S0,#0
	%L1 %R0,%1% ld %S0,#0
	%L1 %R0,%1% ld %S0,#0
	%L1 %R0,%1% ld %S0,#0"
  [(set_attr "length" "2,2,4,4,6,6")])

(define_insn "zero_extendhidi2"
  [(set (match_operand:DI 0 "register_operand" "=x,r,x,r,x,r")
	(zero_extend:DI (match_operand:HI 1 "general_operand" "0,0,?r,?r,??,??g")))
   (clobber (match_scratch:PQI 2 "=X,c,X,c,X,c"))
   (clobber (match_scratch:PQI 3 "=X,c,X,c,X,c"))]
  ""
  "@
	ld %S0,#0% ld %T0,#0% ld %U0,#0
	ld %S0,#0% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% ld %S0,#0% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% ld %S0,#0% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% ld %S0,#0% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% ld %S0,#0% ld %T0,#0% ld %U0,#0"
  [(set_attr "length" "6,6,8,8,10,10")])

;; Splits are necessary to expose clobber of loop cond registers.
(define_insn "zero_extendpsisi2"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=d,r,r,m")
	(zero_extend:SI (match_operand:PSI 1 "general_operand" "a,a,?oi,a")))
   (clobber (match_scratch:HI 2 "=X,&d,X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "@
	ld %R0,%R1% ld %S0,%E1
	ld %2,%E1% ld %R0,%R1% ld %S0,%2
	%L1 %R0,%R1% %L1 %S0,%S1% and %S0,#63
	ldw %0,%1"
  [(set_attr "length" "4,6,12,4")])

(define_split
  [(set (match_operand:SI 0 "nonimmediate_operand" "=d")
	(zero_extend:SI (match_operand:PSI 1 "general_operand" "a")))
   (clobber (match_scratch:HI 2 "=X"))
   (clobber (reg:HI T_REG))]
  "reload_completed
   && (data_reg_operand (operands[0], SImode)
       || memory_operand (operands[1], SImode)
       || immediate_operand (operands[1], SImode))"
  [(pc)]
  "
{
  emit_insn (gen_movhi_i (risc16_split_operand (operands[0], 2, 1),
			  risc16_split_operand (operands[1], 2, 1)));
  emit_insn (gen_movhi_i (risc16_split_operand (operands[0], 2, 0),
			  risc16_split_operand (operands[1], 2, 0)));
  if (! register_operand (operands[1], PSImode))
    emit_insn (gen_andhi3 (risc16_split_operand (operands[0], 2, 0),
			   risc16_split_operand (operands[0], 2, 0),
			   GEN_INT (63)));
  DONE;
}")

(define_split
  [(set (match_operand:SI 0 "arith_reg_operand" "=r")
	(zero_extend:SI (match_operand:PSI 1 "arith_reg_operand" "a")))
   (clobber (match_operand:HI 2 "data_reg_operand" "=&d"))
   (clobber (reg:HI T_REG))]
  "reload_completed"
  [(pc)]
  "
{
  emit_insn (gen_movhi_i (operands[2],
			  risc16_split_operand (operands[1], 2, 0)));
  emit_insn (gen_movhi_i (risc16_split_operand (operands[0], 2, 1),
			  risc16_split_operand (operands[1], 2, 1)));
  emit_insn (gen_movhi_i (risc16_split_operand (operands[0], 2, 0),
			  operands[2]));
  DONE;
}")

(define_insn "zero_extendsidi2"
  [(set (match_operand:DI 0 "register_operand" "=x,r,x,r,x,r")
	(zero_extend:DI (match_operand:SI 1 "general_operand" "0,0,?r,?r,??m,??m")))
   (clobber (match_scratch:PQI 2 "=X,c,X,c,X,c"))
   (clobber (match_scratch:PQI 3 "=X,c,X,c,X,c"))]
  ""
  "@
	ld %T0,#0% ld %U0,#0
	ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% %L1 %S0,%S1% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% %L1 %S0,%S1% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% %L1 %S0,%S1% ld %T0,#0% ld %U0,#0
	%L1 %R0,%R1% %L1 %S0,%S1% ld %T0,#0% ld %U0,#0"
  [(set_attr "length" "4,4,8,8,12,12")])

;;;; Conversions from a larger integer to a smaller integer

;; the d/0 variant doesn't set Z0/Z1 because no insn is needed.
;; ld rn,em doesn't set Z0/Z1 at all, and neither does ldw.
(define_insn "truncsipsi2"
  [(set (match_operand:PSI 0 "nonimmediate_operand" "=d,?a,?a,a,&a,a,?d,m")
	(truncate:PSI (match_operand:SI 1
		       "nonimmediate_operand"        "0,d,0,?a,?a,m,m,?r")))
   (clobber (match_scratch:HI 2 "=X,X,&d,&d,d,X,X,X"))]
  ""
  "@
	; no-op
	ld %R0,%R1% ld %E0,%S1
	ld %2,%S1% ld %E0,%2
	ld %2,%S1% ld %E0,%2% ld %R0,%R1
	ld %R0,%R1% ld %2,%S1% ld %E0,%2
	ldw a%N0,%1
	ldw %R0,%R1% ldw %S0,%S1
	ldw %R0,%R1% ldw %S0,%S1"
  [(set (attr "type") (cond [(eq_attr "load" "yes")
			    (const_string "load")] (const_string "other")))
   (set_attr "length" "0,4,4,6,6,4,8,8")])

;; ::::::::::::::::::::
;; ::
;; :: 16 bit Integer arithmetic
;; ::
;; ::::::::::::::::::::

;; Addition
(define_insn "addhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(plus:HI (match_operand:HI 1 "register_operand" "%0,0")
		 (match_operand:HI 2 "nonmemory_operand" "rKT,i")))
   (clobber (match_scratch:PQI 3 "=X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "@
	%b2 %0,%A2
	add %0,%R2"
  [(set_attr "length" "2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

(define_insn "*add_loopcount"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=j,k,j,k")
	(plus:HI (match_operand:HI 1 "register_operand" "%0,0,0,0")
		 (match_operand:HI 2 "nonmemory_operand" "rKT,rKT,i,i")))
   (set (match_operand:PQI 3 "register_operand" "=v,w,v,w")
	(eq:PQI (plus:HI (match_dup 1) (match_dup 2)) (const_int 0)))
   (clobber (reg:HI T_REG))]
  ""
  "@
	%b2 %0,%A2
	%b2 %0,%A2
	add %0,%R2
	add %0,%R2"
  [(set_attr "length" "2,2,4,4")
   (set_attr "type" "t_set")])

(define_expand "calc_loopcount"
  [(set (match_operand:HI 0 "" "") (match_operand:HI 1 "" ""))
   (set (match_operand:PQI 2 "" "") (eq:PQI (match_dup 1) (const_int 0)))
   (clobber (reg:HI T_REG))]
  ""
  "")

(define_insn "addhic"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(plus:HI (match_operand:HI 1 "register_operand" "0,0,0")
		 (match_operand:HI 2 "nonmemory_operand" "r,KT,i")))
   (set (reg:HI T_REG)
	(ltu:HI (plus:HI (match_dup 1) (match_dup 2)) (match_dup 1)))
   (clobber (match_scratch:PQI 3 "=X,X,X"))]
  ""
  "@
	add %0,%2
	%b2 %0,%A2
	add %0,%R2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Note that the value for a 'true' T_REG is -1.
(define_insn_and_split "adchic"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(plus:HI (match_operand:HI 1 "register_operand" "0,0")
		  (minus:HI (match_operand:HI 2 "nonmemory_operand" "r,i")
			    (reg:HI T_REG))))
   (set (reg:HI T_REG)
	(gtu:HI (plus:SI (zero_extend:SI (match_dup 1))
			 (minus:SI (zero_extend:SI (match_dup 2))
				   (sign_extend:SI (reg:HI T_REG))))
		(const_int 65535)))
   (clobber (match_scratch:PQI 3 "=X,X"))]
  ""
  "@
	adc %0,%2
	adc %0,%R2"
  "operands[2] == const0_rtx
   || GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) == 65535"
  [(pc)]
  "
{
  emit_insn ((*(operands[2] == const0_rtx ? gen_incscc : gen_decscc))
	     (operands[0]));
  DONE;
}"
  [(set_attr "length" "2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

(define_insn "subhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(minus:HI (match_operand:HI 1 "register_operand" "0,0")
		  (match_operand:HI 2 "nonmemory_operand" "rKT,i")))
   (clobber (match_scratch:PQI 3 "=X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "@
	%s2 %0,%A2
	add %0,%M2"
  [(set_attr "length" "2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

(define_insn "*sub_loopcount"
  [(set (match_operand:HI 0 "register_operand" "=j,k,j,k")
	(minus:HI (match_operand:HI 1 "register_operand" "%0,0,0,0")
		  (match_operand:HI 2 "nonmemory_operand" "rKT,rKT,i,i")))
   (set (match_operand:PQI 3 "register_operand" "=v,w,v,w")
	(eq:PQI (minus:HI (match_dup 1) (match_dup 2)) (const_int 0)))
   (clobber (reg:HI T_REG))]
  ""
  "@
	%s2 %0,%A2
	%s2 %0,%A2
	sub %0,%R2
	sub %0,%R2"
  [(set_attr "length" "2,2,4,4")
   (set_attr "type" "t_set")])

;; We can't allow 'i' for operand 3 here because this would permit symbolic
;; constants, that might turn out to be zero at link time.
;; add rn,#0 always clears T, while sub rn,#0 always sets T.
(define_insn "subhic"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(minus:HI (match_operand:HI 1 "register_operand" "0,0,0")
		  (match_operand:HI 2 "nonmemory_operand" "r,KT,n")))
   (set (reg:HI T_REG)
	(leu:HI (minus:HI (match_dup 1) (match_dup 2)) (match_dup 1)))
   (clobber (match_scratch:PQI 3 "=X,X,X"))]
  ""
  "@
	sub %0,%2
	%s2 %0,%A2
	add %0,%M2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Note that the value for a 'true' T_REG is -1.
(define_insn_and_split "sbchic"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(minus:HI (match_operand:HI 1 "register_operand" "0,0")
		  (minus:HI (match_operand:HI 2 "nonmemory_operand" "r,i")
			    (not:HI (reg:HI T_REG)))))
   (set (reg:HI T_REG)
	(ge:HI (minus:SI (zero_extend:SI (match_dup 1))
			 (minus:SI (zero_extend:SI (match_dup 2))
				   (sign_extend:SI (not:HI (reg:HI T_REG)))))
	       (const_int 0)))
   (clobber (match_scratch:PQI 3 "=X,X"))]
  ""
  "@
	sbc %0,%2
	sbc %0,%R2"
  "operands[2] == const0_rtx
   || GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) == 65535"
  [(pc)]
  "
{
  emit_insn ((*(operands[2] == const0_rtx ? gen_decscc : gen_incscc))
	     (operands[0]));
  DONE;
}"
  [(set_attr "length" "2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; The same insn especially constrained for sign extension / scc -
;; We don't want to appear to be using a register that we don't use, or
;; to change T when we actually set it to its old value.
;; Note that the value for a 'true' T_REG is -1.
(define_insn "ldnott"
  [(set (match_operand:HI 0 "register_operand" "=r") (not:HI (reg:HI T_REG)))
   (clobber (match_scratch:PQI 1 "=X"))]
  ""
  "sbc %0,%0"
  [(set_attr "loop_count_set" "yes")])

;; Multiplication (same size)
(define_expand "mulhi3"
  [(set (match_operand:HI 0 "general_operand" "")
	(mult:SI (match_operand:HI 1 "general_operand" "")
		 (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_mulhi3_i ());
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R3_REG)));
  DONE;
}")

(define_insn "mulhi3_i"
  [(set (reg:HI R3_REG) (mult:HI (reg:HI R0_REG) (reg:HI R1_REG)))
   (clobber (reg:HI R0_REG))
   (clobber (reg:HI R1_REG))
   (clobber (reg:HI R2_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___mulhi"
  [(set_attr "type" "sfunc")])

;; Signed multiplication producing 32 bit results from 16 bit inputs
(define_expand "mulhisi3"
  [(set (match_operand:SI 0 "general_operand" "")
	(mult:SI (sign_extend:SI (match_operand:HI 1 "general_operand" ""))
		 (sign_extend:SI (match_operand:HI 2 "general_operand" ""))))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT && !(INTVAL (operands[2]) & ~0xff))
    {
      rtx target = operands[0];

      if (! register_operand (operands[0], SImode))
	target = gen_reg_rtx (SImode);
      operands[1] = force_reg (HImode, operands[1]);
      operands[2] = force_reg (QImode, operands[2]);
      emit_insn (gen_sumulhiqisi3_i (target, operands[1], operands[2]));
      if (target != operands[0])
	emit_move_insn (operands[0], target);
      DONE;
    }
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_mulhisi3_i ());
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R2_REG)));
  DONE;
}")

(define_insn "mulhisi3_i"
  [(set (reg:SI R2_REG)
	(mult:SI (sign_extend:SI (reg:HI R0_REG))
		 (sign_extend:SI (reg:HI R1_REG))))
   (clobber (reg:HI R0_REG))
   (clobber (reg:HI R1_REG))
   (clobber (reg:HI R4_REG))
   (clobber (reg:HI R5_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___smulhisi"
  [(set_attr "type" "sfunc")])

;; Unsigned multiplication producing 16 bit results from 8 bit inputs
(define_insn "umulqihi3"
  [(set (match_operand:HI 0 "register_operand" "=d")
	(mult:HI (zero_extend:HI (match_operand:QI 1 "register_operand" "%0"))
		 (zero_extend:HI (match_operand:QI 2 "nonmemory_operand" "d"))))]
  ""
  "mul uu,%0,%2"
  [(set_attr "length" "2")])

;; Unsigned multiplication producing 32 bit results from 16 bit inputs
(define_expand "umulhisi3"
  [(set (match_operand:SI 0 "general_operand" "")
	(mult:SI (zero_extend:SI (match_operand:HI 1 "general_operand" ""))
		 (zero_extend:SI (match_operand:HI 2 "general_operand" ""))))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT && !(INTVAL (operands[2]) & ~0xff))
    {
      rtx target = operands[0];

      if (! register_operand (operands[0], SImode))
	target = gen_reg_rtx (SImode);
      operands[1] = force_reg (SImode, operands[1]);
      operands[2] = force_reg (QImode, operands[2]);
      emit_insn (gen_umulhiqisi3_i (target, operands[1], operands[2]));
      if (target != operands[0])
	emit_move_insn (operands[0], target);
      DONE;
    }
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_umulhisi3_i ());
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R2_REG)));
  DONE;
}")

(define_insn "umulhisi3_i"
  [(set (reg:SI R2_REG)
	(mult:SI (zero_extend:SI (reg:HI R0_REG))
		 (zero_extend:SI (reg:HI R1_REG))))
   (clobber (reg:HI R1_REG))
   (clobber (reg:HI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___umulhisi"
  [(set_attr "type" "sfunc")])

(define_insn "umulhiqisi3_i"
  [(set (match_operand:SI 0 "register_operand" "=d")
	(mult:SI (zero_extend:SI (match_operand:HI 1 "register_operand" "d"))
		 (zero_extend:SI (match_operand:QI 2 "register_operand" "d"))))
   (clobber (match_scratch:HI 3 "=&d"))
   (clobber (reg:HI T_REG))]
  ""
  "*
{
  if (REGNO (operands[0]) == REGNO (operands[2]))
    {
      output_asm_insn (\"ld %3,%1% srb %3\", operands);
      output_asm_insn (\"mul uu,%3,%2% mul uu,%2,%1\", operands);
      output_asm_insn (\"ld %R0,%2% ld %S0,%3% slb %3\", operands);
    }
  else
    {
      if (REGNO (operands[0]) != REGNO (operands[1]))
	output_asm_insn (\"ld %S0,%1\", operands);
      output_asm_insn (\"ld %3,%1% srb %S0\", operands);
      output_asm_insn (\"mul uu,%S0,%2% mul uu,%3,%2\", operands);
      output_asm_insn (\"ld %R0,%S0% slb %R0\", operands);
    }
  return \"srb %S0% add %R0,%3% incc %S0\";
}"
  [(set_attr "length" "20")
   (set_attr "type" "t_set")])

(define_insn "sumulhiqisi3_i"
  [(set (match_operand:SI 0 "register_operand" "=d")
	(mult:SI (sign_extend:SI (match_operand:HI 1 "register_operand" "d"))
		 (zero_extend:SI (match_operand:QI 2 "register_operand" "d"))))
   (clobber (match_scratch:HI 3 "=&d"))
   (clobber (reg:HI T_REG))]
  ""
  "*
{
  if (REGNO (operands[0]) == REGNO (operands[2]))
    {
      output_asm_insn (\"ld %3,%1% srb %3\", operands);
      output_asm_insn (\"mul su,%3,%2% mul uu,%2,%1\", operands);
      output_asm_insn (\"ld %R0,%2% ld %S0,%3% slb %3\", operands);
    }
  else
    {
      if (REGNO (operands[0]) != REGNO (operands[1]))
	output_asm_insn (\"ld %S0,%1\", operands);
      output_asm_insn (\"ld %3,%1% srb %S0\", operands);
      output_asm_insn (\"mul su,%S0,%2% mul uu,%3,%2\", operands);
      output_asm_insn (\"ld %R0,%S0% slb %R0\", operands);
    }
  return \"srb %S0% add %R0,%3% incc %S0% ext %S0\";
}"
  [(set_attr "length" "22")
   (set_attr "type" "t_set")])

;; Signed Division
(define_expand "divhi3"
  [(set (match_operand:HI 0 "general_operand" "")
	(div:HI (match_operand:HI 1 "general_operand" "")
		(match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_divhi3_i ());
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R0_REG)));
  DONE;
}")

(define_insn "divhi3_i"
  [(set (reg:HI R0_REG) (div:HI (reg:HI R0_REG) (reg:HI R1_REG)))
   (clobber (reg:HI R1_REG))
   (clobber (reg:HI R2_REG))
   (clobber (reg:HI R6_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___sdivhi"
  [(set_attr "type" "sfunc")])

;; Unsigned Division
(define_expand "udivhi3"
  [(set (match_operand:HI 0 "general_operand" "")
	(udiv:HI (match_operand:HI 1 "general_operand" "")
		 (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_udivmodhi_i ());
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R0_REG)));
  DONE;
}")

(define_insn "udivmodhi_i"
  [(set (reg:HI R0_REG) (udiv:HI (reg:HI R0_REG) (reg:HI R1_REG)))
   (set (reg:HI R2_REG) (umod:HI (reg:HI R0_REG) (reg:HI R1_REG)))
   (clobber (reg:HI R6_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___udivmodhi"
  [(set_attr "type" "sfunc")])

;; Unsigned Modulus
(define_expand "umodhi3"
  [(set (match_operand:HI 0 "general_operand" "")
	(umod:HI (match_operand:HI 1 "general_operand" "")
		 (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_udivmodhi_i ());
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R2_REG)));
  DONE;
}")

;; Unsigned division giving both quotient and remainder
(define_expand "udivmodhi4"
  [(set (match_operand:HI 0 "general_operand" "")
	(udiv:SI (match_operand:SI 1 "general_operand" "")
		 (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:HI 3 "general_operand" "")
	(umod:SI (match_dup 1) (match_dup 2)))]
  ""
  "
{
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R1_REG), operands[2]));
  emit_insn (gen_udivmodhi_i ());
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R0_REG)));
  emit_insn (gen_movhi (operands[3], gen_rtx_REG (HImode, R2_REG)));
  DONE;
}")

;; Negation

(define_insn "neghi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(neg:HI (match_operand:HI 1 "register_operand" "0")))
   (clobber (match_scratch:PQI 2 "=X"))
   (clobber (reg:HI T_REG))]
  ""
  "com2 %0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

(define_insn "neghic"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(neg:HI (match_operand:HI 1 "register_operand" "0")))
   (set (reg:HI T_REG) (eq:HI (match_dup 1) (const_int 0)))
   (clobber (match_scratch:PQI 2 "=X"))]
  ""
  "com2 %0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

(define_insn "ngchic"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(minus:HI (not:HI (match_operand:HI 1 "register_operand" "0"))
		  (reg:HI T_REG)))
   (set (reg:HI T_REG)
	(and:HI (eq:HI (match_dup 1) (const_int 0)) (reg:HI T_REG)))
   (clobber (match_scratch:PQI 2 "=X"))]
  ""
  "comc %0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; ::::::::::::::::::::
;; ::
;; :: 22 bit Integer arithmetic
;; ::
;; ::::::::::::::::::::

;; Addition
(define_expand "addpsi3"
  [(set (match_operand:PSI 0 "register_operand" "")
        (plus:PSI (match_operand:PSI 1 "register_operand" "")
                  (match_operand:PSI 2 "nonmemory_operand" "")))]
  ""
  "
{
  if (CONSTANT_P (operands[2]))
    emit_insn (gen_addpsi3_i (operands[0], operands[1], operands[2]));
  else
    {
      rtx si_reg = gen_reg_rtx (SImode);

      emit_insn (gen_zero_extendpsisi2 (si_reg, operands[2]));
      emit_insn (gen_addpsisi (operands[0], operands[1], si_reg));
    }
  DONE;
}")

;; The split is necessary when a loop condition reg is clobbered.
(define_insn_and_split "addpsi3_i"
  [(set (match_operand:PSI 0 "register_operand" "=a,a,a")
	(plus:PSI (match_operand:PSI 1 "register_operand" "%0,0,0")
		  (match_operand:PSI 2 "immediate_operand" "I,MN,i")))
   (clobber (match_scratch:HI 3 "=X,X,&d"))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx low, high;

  if (GET_CODE (operands[2]) == CONST_INT
      && INTVAL (operands[2]) >= -65535
      && INTVAL (operands[2]) <= 65535)
    {
      emit_insn (gen_addpsi_noclobber (operands[0], operands[1], operands[2]));
      DONE;
    }
  /* add %0,%R2\;ld %3,%E0\;add %3,%S2\;ld %E0,%3 */
  low = risc16_split_operand (operands[2], 2, 1);
  high = risc16_split_operand (operands[2], 2, 0);
  if (GET_CODE (high) == CONST_INT && INTVAL (high) & ~63)
    high = GEN_INT (INTVAL (high) & 63);
  emit_label_insn (gen_addpsihi (operands[0], operands[1], low), low);
  emit_insn (gen_movhi_i (operands[3],
			  risc16_split_operand (operands[0], 2, 0)));
  emit_label_insn (gen_addhi3 (operands[3], operands[3], high), high);
  emit_insn (gen_movhi_i (risc16_split_operand (operands[0], 2, 0),
			  operands[3]));
  DONE;
}")

;; Split, because this can clobber a loop condition reg.
(define_insn_and_split "addpsisi"
  [(set (match_operand:PSI 0 "register_operand" "=&a,a")
	(plus:PSI (match_operand:PSI 1 "register_operand" "0,0")
		  (truncate:PSI (match_operand:SI 2 "nonmemory_operand" "r,i"))))
   (clobber (match_scratch:HI 3 "=&d,d"))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  /* add %0,%R2% ld %3,%E0% add %3,(%S2 & 0x3f)% ld %E0,%3 */
  rtx low2 = risc16_split_operand (operands[2], 2, 1);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  if (GET_CODE (high2) == CONST_INT && INTVAL (high2) & ~63)
    high2 = GEN_INT (INTVAL (high2) & 63);
  if (low2 != const0_rtx)
    emit_label_insn (gen_addpsihi (operands[0], operands[0], low2), low2);
  if (high2 != const0_rtx)
    {
      emit_insn (gen_movhi_i (operands[3], high0));
      emit_label_insn (gen_addhi3 (operands[3], operands[3], high2), high2);
      emit_insn (gen_movhi_i (high0, operands[3]));
    }
  DONE;
}")

(define_insn "addpsihi"
  [(set (match_operand:PSI 0 "register_operand" "=a,a")
	(plus:PSI (match_operand:PSI 1 "register_operand" "0,0")
		  (zero_extend:PSI (match_operand:HI 2
				    "nonmemory_operand" "Ir,?i"))))]
  ""
  "%b2 %0,%A2"
  [(set_attr "length" "2,4")])

;; Split is necessary because of loopcond clobber
(define_insn_and_split "addsipsi"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "0")
		 (zero_extend:SI (match_operand:PSI 2 "register_operand" "a"))))
   (clobber (match_scratch:HI 3 "=&d"))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; ld %3,%E2% add %R0,%R2% add %S0,%3
  "reload_completed"
  [(pc)]
  "
{
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);
  rtx low0 = risc16_split_operand (operands[0], 2, 1);
  rtx low2 = risc16_split_operand (operands[2], 2, 1);

  emit_insn (gen_movhi_i (operands[3], high2));
  emit_insn (gen_addhic (low0, low0, low2));
  emit_insn (gen_adchic (high0, high0, operands[3]));
  DONE;
}")

;; combiner pattern
(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(zero_extend:SI (plus:PSI (truncate:PSI (zero_extend:SI (match_operand:PSI 1 "register_operand" "")))
				  (match_operand:PSI 2 "register_operand" ""))))]
  "rtx_equal_p (operands[1], operands[2])"
  [(pc)]
  "
{
  emit_insn (gen_zero_extendpsisi2 (operands[0], operands[1]));
  emit_insn (gen_addsi_trunc (operands[0], operands[0], operands[0]));
  DONE;
}")

;; Split is necessary to handle loopcond clobber
(define_insn_and_split "addsi_trunc"
  [(set (match_operand:SI 0 "register_operand" "=r,&r")
	(and:SI (plus:SI (match_operand:SI 1 "register_operand" "0,0")
			 (match_operand:SI 2 "register_operand" "0,r"))
		(const_int 4194303)))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; add %R0,%R1% adc %S0,%S1% and %S0,#63
  "reload_completed"
  [(pc)]
  "
{
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);
  rtx low0 = risc16_split_operand (operands[0], 2, 1);
  rtx low2 = risc16_split_operand (operands[2], 2, 1);

  emit_insn (gen_addhic (low0, low0, low2));
  emit_insn (gen_adchic (high0, high0, high2));
  emit_insn (gen_andhi3 (high0, high0, GEN_INT (63)));
  DONE;
}")

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(ashift:SI (zero_extend:SI (truncate:PSI (match_operand:SI 1 "register_operand" "")))
		   (const_int 1)))]
  ""
  [(pc)]
  "
{
  emit_insn (gen_addsi3 (operands[0], operands[1], operands[1]));
  emit_insn (gen_andsi3 (operands[0], operands[0], GEN_INT (0x7fffff)));
  DONE;
}")

(define_expand "stack_adjust"
  [(parallel [(set (match_operand:PSI 0 "" "")
	      (plus:PSI (match_dup 0) (match_operand:PSI 1 "" "")))
	      (clobber (match_operand:HI 2 "" ""))
	      (clobber (reg:PQI T_REG))])]
  ""
  "")

;; special patterns for address reloads
(define_insn "addpsi_leavet"
  [(set (match_operand:PSI 0 "register_operand" "=a,a,a,a,&a")
	(plus:PSI (match_operand:PSI 1 "register_operand" "%0,0,0,0,a")
		  (match_operand:PSI 2 "nonmemory_operand" "I,d,MN,O,dO")))
   (clobber (match_scratch:PSI 3 "=X,&a,X,&a,X"))]
  "reload_in_progress || reload_completed"
  "@
	%b2 %0,%A2
	add %0,%R2\;ld %R3,%E0\;add %3,%S2\;ld %E0,%R3
	%b2 %0,%A2
	%b2 %0,%A2\;ld %R3,%E0\;%b2 %3,%B2\;ld %E0,%R3
	ld %R0,%E1\;%b2 %0,%B2\;ld %E0,%R0\;ld %R0,%R1\;%b2 %0,%A2"
  [(set_attr "length" "2,8,4,10,12")])

(define_insn "addpsi_noclobber"
  [(set (match_operand:PSI 0 "register_operand" "=a,a,a,a,b,b")
	(plus:PSI (match_operand:PSI 1 "register_operand" "%0,0,?a,??d,a,d")
		  (match_operand:PSI 2 "nonmemory_operand" "I,?MN,?MN,?MN,!i,!i")))]
  "reload_in_progress || reload_completed"
  "@
	%b2 %0,%A2
	%b2 %0,%A2
	ld %0,%1\;%b2 %0,%A2
	ld %E0,%S1\;ld %R0,%R1\;%b2 %0,%A2
	ld %0,%1\;add %0,%R2\;push r0,r1\;rrc r0\;ld r1,%E0\;add r1,%S2\;add r0,r0\;ld %E0,r1\;pop r1,r0
	push r0,r1\;rrc r0\;ld %R0,%R1\;add %R0,%R2\;ld r1,%S1\;adc r1,%S2\;add r0,r0\;ld %E0,r1\;pop r1,r0"
  [(set_attr "length" "2,4,6,8,20,20")])

(define_peephole2
  [(set (match_operand:PSI 0 "register_operand" "=a")
	(plus:PSI (match_operand:PSI 1 "register_operand" "a")
		  (match_operand:PSI 2 "immediate_operand" "i")))
   (match_scratch:HI 3 "d")]
  "(GET_CODE (operands[2]) != CONST_INT
    || INTVAL (operands[2]) < -65535
    || INTVAL (operands[2]) > 65535)
   && peep2_regno_dead_p (1, T_REG)"
  [(parallel [(set (match_operand:PSI 0 "register_operand" "")
		   (plus:PSI (match_dup 0) (match_dup 2)))
	      (clobber (match_dup 3))
	      (clobber (reg:PQI T_REG))])]
  "
{
  if (! rtx_equal_p (operands[0], operands[1]))
    emit_insn (gen_movpsi_i (operands[0], operands[1]));
}")

;; Subtraction
(define_expand "subpsi3"
  [(set (match_operand:PSI 0 "register_operand" "")
	(minus:PSI (match_operand:PSI 1 "register_operand" "")
		   (match_operand:PSI 2 "nonmemory_operand" "")))]
  ""
  "
{
  if (CONSTANT_P (operands[2]))
    {
      rtx const1, const2;

      const1 = operands[2];
      if (GET_CODE (const1) == CONST)
	const1 = XEXP (const1, 0);
      const1 = gen_rtx_NEG (PSImode, const1);
      const2 = simplify_rtx (const1);
      if (const2)
	const1 = const2;
      if (! CONSTANT_P (const1))
	const1 = gen_rtx_CONST (PSImode, const1);

      emit_insn (gen_addpsi3_i (operands[0], operands[1], const1));
    }
  else
    {
      rtx si_reg = gen_reg_rtx (SImode);

      emit_insn (gen_zero_extendpsisi2 (si_reg, operands[2]));
      emit_insn (gen_subpsisi (operands[0], operands[1], si_reg));
    }
  DONE;
}")

;; Split, because this can clobber a loop condition reg.
(define_insn_and_split "subpsisi"
  [(set (match_operand:PSI 0 "register_operand" "=a,a")
	(minus:PSI (match_operand:PSI 1 "register_operand" "0,0")
		   (truncate:PSI (match_operand:SI 2 "register_operand" "r,i"))))
   (clobber (match_scratch:HI 3 "=&d,d"))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  /* sub %0,%R2% ld %3,%E0% sub %3,(%S2 & 0x3f)% ld %E0,%3 */
  rtx low2 = risc16_split_operand (operands[2], 2, 1);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  if (GET_CODE (high2) == CONST_INT && INTVAL (high2) & ~63)
    high2 = GEN_INT (INTVAL (high2) & 63);
  if (low2 != const0_rtx)
    emit_label_insn (gen_subpsihi (operands[0], operands[0], low2), low2);
  if (high2 != const0_rtx)
    {
      emit_insn (gen_movhi_i (operands[3], high0));
      emit_label_insn (gen_subhi3 (operands[3], operands[3], high2), high2);
      emit_insn (gen_movhi_i (high0, operands[3]));
    }
  DONE;
}")

(define_insn "subpsihi"
  [(set (match_operand:PSI 0 "register_operand" "=a,a")
	(minus:PSI (match_operand:PSI 1 "register_operand" "0,0")
		   (zero_extend:PSI (match_operand:HI 2
				     "nonmemory_operand" "Ir,?MN"))))]
  ""
  "%s2 %0,%A2"
  [(set_attr "length" "2,4")])

;; combiner patterns, used for dynamic allocation on stack
(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(zero_extend:SI (truncate:PSI (match_operand:SI 1 "register_operand" ""))))]
  ""
  [(set (match_dup 0) (and:SI (match_dup 1) (const_int 0x3fffff)))]
  "")

(define_split
  [(set (match_operand:PSI 0 "register_operand" "")
	(minus:PSI (match_operand:PSI 1 "register_operand" "")
		   (truncate:PSI (zero_extend:SI
		   (plus:PSI (match_operand:PSI 2 "register_operand" "")
			     (truncate:PSI (match_operand:SI 3 "register_operand" "")))))))
   (clobber (match_operand:SI 4 "register_operand" ""))]
  ""
  [(pc)]
  "
{
  emit_insn (gen_addsipsi (operands[4], operands[3], operands[2]));
  emit_insn (gen_subpsisi (operands[0], operands[1], operands[4]));
  DONE;
}")

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(ashift:SI (zero_extend:SI (ashiftrt:PSI (match_operand:PSI 1 "register_operand" "")
						 (const_int 1)))
		   (const_int 1)))]
  ""
  [(pc)]
  "
{
  emit_insn (gen_zero_extendpsisi2 (operands[0], operands[1]));
  emit_insn (gen_andsi3 (operands[0], operands[0], GEN_INT (~1)));
  DONE;
}")

(define_split
  [(set (match_operand:SI 0 "register_operand" "")
	(zero_extend:SI (plus:PSI (zero_extend:PSI (match_operand:HI 1 "register_operand" ""))
				  (match_operand:SI 2 "immediate_operand" ""))))]
  ""
  [(pc)]
  "
{
  emit_insn (gen_zero_extendhisi2 (operands[0], operands[1]));
  emit_insn (gen_addsi3 (operands[0], operands[0], operands[2]));
  DONE;
}")

;; ::::::::::::::::::::
;; :
;; :: 32 bit Integer arithmetic
;; ::
;; ::::::::::::::::::::

;; Addition
(define_insn_and_split "addsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(plus:SI (match_operand:SI 1 "register_operand" "%0,0")
		 (match_operand:SI 2 "nonmemory_operand" "r,i")))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx low0 = risc16_split_operand (operands[0], 2, 1);
  rtx low2 = risc16_split_operand (operands[2], 2, 1);
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);

  if (low2 == const0_rtx)
    {
      emit_insn (gen_addhi3 (high0, high0, high2));
      DONE;
    }
  emit_insn (gen_addhic (low0, low0, low2));
  emit_insn (gen_adchic (high0, high0, high2));
  DONE;
}")

;; Subtraction
(define_insn_and_split "subsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(minus:SI (match_operand:SI 1 "register_operand" "0,0")
		  (match_operand:SI 2 "nonmemory_operand" "r,i")))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx low0 = risc16_split_operand (operands[0], 2, 1);
  rtx low2 = risc16_split_operand (operands[2], 2, 1);
  rtx high0 = risc16_split_operand (operands[0], 2, 0);
  rtx high2 = risc16_split_operand (operands[2], 2, 0);

  if (low2 == const0_rtx)
    {
      emit_insn (gen_subhi3 (high0, high0, high2));
      DONE;
    }
  emit_insn (gen_subhic (low0, low0, low2));
  emit_insn (gen_sbchic (high0, high0, high2));
  DONE;
}")

;; Used by casesi.
(define_insn "subsic"
  [(set (match_operand:SI 0 "register_operand" "=x,r,x,r,x,r,x,r")
	(minus:SI (match_operand:SI 1 "register_operand" "0,0,0,0,0,0,0,0")
		  (match_operand:SI 2 "nonmemory_operand" "r,r,K,K,MN,MN,i,i")))
   (set (reg:PQI T_REG) (leu:PQI (match_dup 2) (match_dup 0)))
   (clobber (match_scratch:PQI 3 "=X,&c,X,&c,X,&c,X,&c"))
   (clobber (match_scratch:PQI 4 "=X,&c,X,&c,X,&c,X,&c"))]
  ""
  "@
	sub %R0,%R2\;sbc %S0,%S2
	sub %R0,%R2\;sbc %S0,%S2
	%s2 %R0,%A2\;%d2 %S0
	%s2 %R0,%A2\;%d2 %S0
	sub %R0,%R2\;%d2 %S0
	sub %R0,%R2\;%d2 %S0
	sub %R0,%R2\;sbc %S0,%S2
	sub %R0,%R2\;sbc %S0,%S2"
  [(set_attr "length" "4,4,4,4,6,6,8,8")
   (set_attr "type" "t_set")])

;; Multiplication (same size)
(define_expand "mulsi3"
  [(set (match_operand:SI 0 "general_operand" "")
	(mult:SI (match_operand:SI 1 "general_operand" "")
		 (match_operand:SI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT && ! (INTVAL (operands[2]) & 0xffff))
    {
      rtx reg = force_reg (SImode, operands[0]);
      rtx target = gen_rtx_SUBREG (HImode, reg, 0), result;

      if (! reg_overlap_mentioned_p (operands[0], operands[1]))
	emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
      
      result = expand_mult (HImode,
			    operand_subword (operands[1], 1, 0, SImode),
			    operand_subword (operands[2], 0, 0, SImode),
			    target, 1);
      if (result != target)
	emit_move_insn (target, result);
      emit_insn (gen_movhi_i (gen_rtx_SUBREG (HImode, reg, 1), const0_rtx));
      emit_move_insn (reg, reg);
      DONE;
    }
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT && ! (INTVAL (operands[2]) & -256))
    {
      emit_insn (gen_movqi (gen_rtx_REG (QImode, R3_REG), operands[2]));
      emit_insn (gen_mulsiqisi_i ());
      emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R0_REG)));
    }
  else
    {
      emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[2]));
      emit_insn (gen_mulsi3_i ());
      emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R10_REG)));
    }
  DONE;
}")

(define_insn "mulsi3_i"
  [(set (reg:SI R10_REG) (mult:SI (reg:SI R0_REG) (reg:SI R2_REG)))
   (clobber (reg:SI R0_REG))
   (clobber (reg:HI R6_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___mulsi"
  [(set_attr "type" "sfunc")])

(define_insn "mulsiqisi_i"
  [(set (reg:SI R0_REG)
	(mult:SI (reg:SI R0_REG) (zero_extend:SI (reg:QI R3_REG))))
   (clobber (reg:HI R6_REG))
   (clobber (reg:HI R10_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___mulsiqisi"
  [(set_attr "type" "sfunc")])

;; Unsigned multiplication producing 64 bit results from 32 bit inputs
(define_expand "umulsidi3"
  [(set (match_operand:DI 0 "general_operand" "")
	(mult:DI (zero_extend:DI (match_operand:SI 1 "general_operand" ""))
		 (zero_extend:DI (match_operand:SI 2 "general_operand" ""))))]
  ""
  "
{
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[2]));
  emit_insn (gen_umulsidi3_i ());
  emit_insn (gen_movdi (operands[0], gen_rtx_REG (DImode, R8_REG)));
  DONE;
}")

(define_insn "umulsidi3_i"
  [(set (reg:DI R8_REG)
	(mult:DI (zero_extend:DI (reg:SI R0_REG))
		 (zero_extend:DI (reg:SI R2_REG))))
   (clobber (reg:SI R0_REG))
   (clobber (reg:HI R3_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PQI Z1_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___umulsidi"
  [(set_attr "type" "sfunc")])

;; Signed Division
(define_expand "divsi3"
  [(set (match_operand:SI 0 "general_operand" "")
	(div:SI (match_operand:SI 1 "general_operand" "")
		(match_operand:SI 2 "general_operand" "")))]
  ""
  "
{
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[2]));
  emit_insn (gen_divsi3_i ());
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R0_REG)));
  DONE;
}")

(define_insn "divsi3_i"
  [(set (reg:SI R0_REG) (div:SI (reg:SI R0_REG) (reg:SI R2_REG)))
   (clobber (reg:SI R2_REG))
   (clobber (reg:SI R5_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___sdivsi"
  [(set_attr "type" "sfunc")])

;;;; Unsigned Division
;;(define_expand "udivsi3"
;;  [(set (match_operand:SI 0 "general_operand" "")
;;	(udiv:SI (match_operand:SI 1 "general_operand" "")
;;		 (match_operand:SI 2 "general_operand" "")))]
;;  ""
;;  "
;;{
;;  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
;;  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[2]));
;;  emit_insn (gen_udivmodsi_i ());
;;  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R0_REG)));
;;  DONE;
;;}")

(define_insn "udivmodsi_i"
  [(set (reg:SI R0_REG) (udiv:SI (reg:SI R0_REG) (reg:SI R2_REG)))
   (set (reg:SI R2_REG) (umod:SI (reg:SI R0_REG) (reg:SI R2_REG)))
   (clobber (reg:SI R5_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___udivmodsi"
  [(set_attr "type" "sfunc")])

;;;; Unsigned Modulus
;;(define_expand "umodsi3"
;;  [(set (match_operand:SI 0 "general_operand" "")
;;	(umod:SI (match_operand:SI 1 "general_operand" "")
;;		 (match_operand:SI 2 "general_operand" "")))]
;;  ""
;;  "
;;{
;;  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
;;  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[2]));
;;  emit_insn (gen_udivmodsi_i ());
;;  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R2_REG)));
;;  DONE;
;;}")

;; Unsigned division giving both quotient and remainder
;;
;; We want div & mod to be combined - e.g. for divdi3.
;; Unfortunately, this doesn't work when they are explicitly using
;; the hard registers that they use.
;; So we use a pattern that pretends to be able to operate on general operands,
;; and clobber r0..r3.  Then we provide a splitter to break this up into
;; the proper insns.  With sched1 enabled, this will actually be split before
;; register allocation, thus not impeding register allocation.
(define_insn_and_split "udivmodsi4"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=g")
	(udiv:SI (match_operand:SI 1 "general_operand" "g")
		 (match_operand:SI 2 "general_operand" "g")))
   (set (match_operand:SI 3 "nonimmediate_operand" "=g")
	(umod:SI (match_dup 1) (match_dup 2)))
   (clobber (reg:SI R0_REG))
   (clobber (reg:SI R2_REG))
   (clobber (reg:SI R5_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "#"
  ""
  [(pc)]
  "
{
  rtx r0 = gen_rtx_REG (SImode, R0_REG);
  rtx r2 = gen_rtx_REG (SImode, R2_REG);

  if (reg_overlap_mentioned_p (r0, operands[2]))
    {
      if (reg_overlap_mentioned_p (r2, operands[1]))
	{
	  rtx tmp;

	  if (no_new_pseudos)
	    tmp = gen_rtx_REG (SImode, R5_REG);
	  else
	    tmp = gen_reg_rtx (SImode);
	  emit_insn (gen_movsi (tmp, operands[1]));
	  operands[1] = tmp;
	}
      emit_insn (gen_movsi (r2, operands[2]));
      operands[2] = r2;
    }
  if (! rtx_equal_p (r0, operands[1]))
    emit_insn (gen_movsi (r0, operands[1]));
  if (! rtx_equal_p (r2, operands[2]))
    emit_insn (gen_movsi (r2, operands[2]));

  emit_insn (gen_udivmodsi_i ());

  emit_insn (gen_movsi (operands[0], r0));
  emit_insn (gen_movsi (operands[3], r2));
  DONE;
}")

;; Negation

(define_insn_and_split "negsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(neg:SI (match_operand:SI 1 "register_operand" "0")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx high = risc16_split_operand (operands[0], 2, 0);
  rtx low = risc16_split_operand (operands[0], 2, 1);

  emit_insn (gen_neghic (low, low));
  emit_insn (gen_ngchic (high, high));
  DONE;
}")

;; ::::::::::::::::::::
;; ::
;; :: 64 bit Integer arithmetic
;; ::
;; ::::::::::::::::::::

;; Addition
;; ??? &r/0/r0 confuses reload, it then RELOAD_OTHERs op0 & 2 and forgets
;; about op1.  So use separate alternatives.
(define_insn_and_split "adddi3"
  [(set (match_operand:DI 0 "register_operand" "=&r,r,r,&r")
	(plus:DI (match_operand:DI 1 "register_operand" "%0,0,0,0")
		 (match_operand:DI 2 "general_operand"  "r,0,n,r0m")))
   (clobber (match_scratch:HI 3 "=X,X,X,&r"))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx op0, op2;
  int word;
  rtx (*gen_fn) PARAMS ((rtx, rtx, rtx)) = gen_addhic;

  for (word = 3; word >= 0; word--)
    {
      op0 = risc16_split_operand (operands[0], 4, word);
      op2 = risc16_split_operand (operands[2], 4, word);
      if (GET_CODE (op2) == MEM)
	{
	  emit_insn (gen_movhi_i (operands[3], op2));
	  op2 = operands[3];
	}
      else if (gen_fn == gen_addhic && op2 == const0_rtx)
	continue;
      else if (gen_fn == gen_addhic && GET_CODE (op2) == CONST_INT
	       && INTVAL (op2) >= 65536 - 127)
	{
	  gen_fn = gen_subhic;
	  op2 = GEN_INT (65536 - INTVAL (op2));
	}
      emit_insn ((*gen_fn) (op0, op0, op2));
      gen_fn = gen_adchic;
    }
  DONE;
}")

;; Subtraction
(define_insn_and_split "subdi3"
  [(set (match_operand:DI 0 "register_operand" "=&r,r,r,&r")
	(minus:DI (match_operand:DI 1 "register_operand" "0,0,0,0")
		  (match_operand:DI 2 "general_operand"  "r,0,n,r0m")))
   (clobber (match_scratch:HI 3 "=X,X,X,&r"))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx op0, op2;
  int word;
  rtx (*gen_fn) PARAMS ((rtx, rtx, rtx)) = gen_subhic;

  for (word = 3; word >= 0; word--)
    {
      op0 = risc16_split_operand (operands[0], 4, word);
      op2 = risc16_split_operand (operands[2], 4, word);
      if (GET_CODE (op2) == MEM)
	{
	  emit_insn (gen_movhi_i (operands[3], op2));
	  op2 = operands[3];
	}
      else if (gen_fn == gen_subhic && op2 == const0_rtx)
	continue;
      else if (gen_fn == gen_subhic && GET_CODE (op2) == CONST_INT
	       && INTVAL (op2) >= 65536 - 127)
	{
	  gen_fn = gen_addhic;
	  op2 = GEN_INT (65536 - INTVAL (op2));
	}
      emit_insn ((*gen_fn) (op0, op0, op2));
      gen_fn = gen_sbchic;
    }
  DONE;
}")

;; Negation
(define_insn_and_split "negdi2"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(neg:DI (match_operand:DI 1 "register_operand" "0")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "
{
  rtx word = risc16_split_operand (operands[0], 2, 3);
  int i;

  emit_insn (gen_neghic (word, word));
  for (i = 2; i >= 0; i--)
    {
      word = risc16_split_operand (operands[0], 2, i);
      emit_insn (gen_ngchic (word, word));
    }
  DONE;
}")

;; ::::::::::::::::::::
;; ::
;; :: 16 bit Integer Shifts and Rotates
;; ::
;; ::::::::::::::::::::

;; Arithmetic Shift Left
(define_expand "ashlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(ashift:HI (match_operand:HI 1 "general_operand" "")
		   (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4
	  || INTVAL (operands[2]) >= 8 && INTVAL (operands[2]) < 11))
    {
      rtx reg;
      int i;

      reg = GET_CODE (operands[0]) == REG ? operands[0] : gen_reg_rtx (HImode);
      if (! rtx_equal_p (reg, operands[1])
	  || (GET_CODE (operands[0]) == REG && operands[2] == const0_rtx))
	emit_insn (gen_movhi (reg, operands[1]));
      i = INTVAL (operands[2]);
      if (i >= 8)
	{
	  emit_insn (gen_ashlhi3_8 (reg));
	  i -= 8;
	}
      while (--i >= 0)
	emit_insn (gen_addhi3 (reg, reg, reg));
      if (reg != operands[0])
	emit_insn (gen_movhi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_ashlhi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashlhi3_n ());
    }
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R0_REG)));
  DONE;
}")

(define_insn "ashlhi3_n"
  [(set (reg:HI R0_REG) (ashift:HI (reg:HI R0_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashlhi_n"
  [(set_attr "type" "sfunc")])

(define_insn "ashlhi3_c"
  [(set (reg:HI R0_REG)
	(ashift:HI (reg:HI R0_REG) (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "* return INTVAL (operands[0]) > 15U ? \"\" : \"jsr ___ashlhi_%O0\";"
  [(set_attr "type" "sfunc")])

(define_insn "ashlhi3_8"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(ashift:HI (match_dup 0) (const_int 8)))
   (clobber (reg:HI T_REG))]
  ""
  "slb %0"
  [(set_attr "type" "t_set")])

;; ??? synth_mult seems not to do much good anyways, because it doesn't
;; take the mode into account when estimating the cost of a shift.
;;;; init_expmed relies on being able to recognize a shift pattern with pseudos.
;;(define_insn "ashlhi3_fake"
;;  [(set (match_operand:HI 0 "register_operand" "+z")
;;	(ashift:HI (match_dup 0) (match_operand:HI 1 "const_int_operand" "n")))
;;   (clobber (reg:HI T_REG))
;;   (clobber (reg:PSI LINK_REG))]
;;  "INTVAL (operands[1]) & 7"
;;  "* return INTVAL (operands[1]) > 15U ? \"\" : \"jsr ___ashlhi_%O1\";"
;;  [(set_attr "type" "sfunc")])

;; Arithmetic Shift Right
(define_expand "ashrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(ashiftrt:HI (match_operand:HI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 4)
    {
      rtx reg;
      int i;

      reg = GET_CODE (operands[0]) == REG ? operands[0] : gen_reg_rtx (HImode);
      emit_insn (gen_movhi (reg, operands[1]));
      for (i = INTVAL (operands[2]); --i >= 0;)
	emit_insn (gen_ashrhi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movhi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R3_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_ashrhi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashrhi3_n ());
    }
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R3_REG)));
  DONE;
}")

(define_insn "ashrhi3_n"
  [(set (reg:HI R3_REG) (ashiftrt:HI (reg:HI R3_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashrhi_n"
  [(set_attr "type" "sfunc")])

(define_insn "ashrhi3_c"
  [(set (reg:HI R3_REG)
	(ashiftrt:HI (reg:HI R3_REG)
		     (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "* return INTVAL (operands[0]) > 15U ? \"\" : \"jsr ___ashrhi_%O0\";"
  [(set_attr "type" "sfunc")])


(define_insn "ashrhi3_1"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(ashiftrt:HI (match_dup 0) (const_int 1)))
   (clobber (reg:HI T_REG))]
  ""
  "sra %0"
  [(set_attr "type" "t_set")])

;; Logical Shift Right
(define_expand "lshrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(lshiftrt:HI (match_operand:HI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4
	  || INTVAL (operands[2]) >= 8 && INTVAL (operands[2]) < 11))
    {
      rtx reg;
      int i;

      reg = GET_CODE (operands[0]) == REG ? operands[0] : gen_reg_rtx (HImode);
      emit_insn (gen_movhi (reg, operands[1]));
      i = INTVAL (operands[2]);
      if (i >= 8)
	{
	  emit_insn (gen_lshrhi3_8 (reg));
	  i -= 8;
	}
      while (--i >= 0)
	emit_insn (gen_lshrhi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movhi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movhi (gen_rtx_REG (HImode, R3_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_lshrhi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_lshrhi3_n ());
    }
  emit_insn (gen_movhi (operands[0], gen_rtx_REG (HImode, R3_REG)));
  DONE;
}")

(define_insn "lshrhi3_n"
  [(set (reg:HI R3_REG) (lshiftrt:HI (reg:HI R3_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___lshrhi_n"
  [(set_attr "type" "sfunc")])

(define_insn "lshrhi3_c"
  [(set (reg:HI R3_REG)
	(lshiftrt:HI (reg:HI R3_REG) (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "* return INTVAL (operands[0]) > 15U ? \"\" : \"jsr ___lshrhi_%O0\";"
  [(set_attr "type" "sfunc")])

(define_insn "lshrhi3_1"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(lshiftrt:HI (match_dup 0) (const_int 1)))
   (clobber (reg:HI T_REG))]
  ""
  "sr %0"
  [(set_attr "type" "t_set")])

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "lshrhic"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(lshiftrt:HI (match_dup 0) (const_int 1)))
   (set (reg:HI T_REG) (neg:HI (and:HI (match_dup 0) (const_int 1))))]
  ""
  "sr %0"
  [(set_attr "type" "t_set")])

(define_insn "lshrhi3_8"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(lshiftrt:HI (match_dup 0) (const_int 8)))
   (clobber (reg:HI T_REG))]
  ""
  "srb %0"
  [(set_attr "type" "t_set")])

;; Rotate Left
(define_expand "rotlhi3"
  [(set (match_operand:HI 0 "register_operand" "")
	(rotate:HI (match_operand:HI 1 "register_operand" "")
		   (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotlhi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn "rotlhi3_1"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(rotate:HI (match_operand:HI 1 "register_operand" "0")
		   (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "rl %0"
  [(set_attr "type" "t_set")])

;; Rotate Right
(define_expand "rotrhi3"
  [(set (match_operand:HI 0 "register_operand" "")
	(rotatert:HI (match_operand:HI 1 "register_operand" "")
		     (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotrhi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn "rotrhi3_1"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(rotatert:HI (match_operand:HI 1 "register_operand" "0")
		     (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "rr %0"
  [(set_attr "type" "t_set")])


;; ::::::::::::::::::::
;; ::
;; :: 22 bit Integer Shifts and Rotates
;; ::
;; ::::::::::::::::::::

;; ??? There is really no sane reason to divide / shift right a pointer,
;; but round_push uses divide / multiply to round the size being allocated
;; on the stack (e.g. alloca of variable size).  We want this to be
;; collapsed into an andpsi insn, so we must not expand lshrpsi into
;; lots of separate insns before combine can have a got at it.
;; combine is only enabled when optimizing, so only provide this pattern
;; in this case.

(define_expand "lshrpsi3"
  [(parallel [(set (match_operand:PSI 0 "general_operand" "")
		   (lshiftrt:PSI (match_operand:PSI 1 "general_operand" "")
				 (match_operand:HI 2 "const_1_operand" "")))
	      (clobber (reg:PQI T_REG))])]
  ""
  "
{
  if (1 || ! optimize)
    {
      rtx si_reg = gen_reg_rtx (SImode);
      emit_insn (gen_zero_extendpsisi2 (si_reg, operands[1]));
      emit_insn (gen_lshrsi3 (si_reg, si_reg, operands[2]));
      emit_insn (gen_truncsipsi2 (operands[0], si_reg));
      DONE;
    }
  operands[0] = force_reg (PSImode, operands[0]);
  if (operands[1] != operands[0])
    emit_move_insn (operands[0], operands[1]);
  operands[1] = operands[0];
}")

(define_insn "lshrpsi3_1"
  [(set (match_operand:PSI 0 "register_operand" "+r")
	(lshiftrt:PSI (match_dup 0) (const_int 1)))
   (clobber (reg:PQI T_REG))]
  "optimize"
  "#")

;; Logical Shift Right

;; ::::::::::::::::::::
;; ::
;; :: 32 bit Integer Shifts and Rotates
;; ::
;; ::::::::::::::::::::

;; Arithmetic Shift Left
(define_expand "ashlsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
	(ashift:SI (match_operand:SI 1 "general_operand" "")
		   (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4 || (INTVAL (operands[2]) >= 16)))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (SImode);
      i = INTVAL (operands[2]);
      if (i >= 16)
	{
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_ashlhi3 (gen_highpart (HImode, reg),
				  gen_lowpart (HImode, operands[1]),
				  GEN_INT (i - 16)));
	  emit_insn (gen_movhi (gen_lowpart (HImode, reg), const0_rtx));
	  /* Emit the last move even if it's a no-op move, so that we
	     have something to tag REG_EQUAL notes on.  */
	  emit_insn (gen_movsi (operands[0], reg));
	  DONE;
	}
      else
	emit_insn (gen_movsi (reg, operands[1]));
      while (--i >= 0)
	emit_insn (gen_addsi3 (reg, reg, reg));
      if (reg != operands[0])
	emit_insn (gen_movsi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 16)
    emit_insn (gen_ashlsi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashlsi3_n ());
    }
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R0_REG)));
  DONE;
}")

(define_insn "ashlsi3_n"
  [(set (reg:SI R0_REG) (ashift:SI (reg:SI R0_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:HI R4_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashlsi_n"
  [(set_attr "type" "sfunc")])

;;(define_insn_and_split "ashlsi3_c_pseudo"
;;  [(set (match_operand:SI 0 "general_operand" "=Z")
;;	(ashift:SI (match_operand:SI 1 "general_operand" "0")
;;		   (match_operand:HI 2 "const_int_operand" "n")))
;;   (clobber (reg:SI R0_REG))
;;   (clobber (reg:HI T_REG))
;;   (clobber (reg:PSI A8_REG))
;;   (clobber (reg:PSI LINK_REG))]
;;  "INTVAL (operands[2]) < 16U && 0"
;;  "#"
;;  ""
;;  [(pc)]
;;  "
;;{
;;  rtx reg0 = gen_rtx_REG (SImode, R0_REG);
;;
;;  emit_move_insn (reg0, operands[1]);
;;  emit_insn (gen_ashlsi3_c (operands[2]));
;;  emit_move_insn (operands[0], reg0);
;;  DONE;
;;}")

(define_insn "ashlsi3_c"
  [(set (reg:SI R0_REG)
	(ashift:SI (reg:SI R0_REG)
		   (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 16U"
  "jsr ___ashlsi_%O0"
  [(set_attr "type" "sfunc")])

;; Arithmetic Shift Right
(define_expand "ashrsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
	(ashiftrt:SI (match_operand:SI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4 || (INTVAL (operands[2]) >= 16)))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (SImode);
      i = INTVAL (operands[2]);
      if (i >= 16)
	{
	  /* ??? Can't use gen_highpart here because it refuses to work on
	     SUBREGs with SUBREG_WORD != 0.  */
	  emit_insn (gen_extendhisi2 (reg, operand_subword (operands[1], 0, 0,
							    SImode)));
	  emit_insn (gen_ashrhi3 (gen_lowpart (HImode, reg),
				  gen_lowpart (HImode, reg),
				  GEN_INT (i - 16)));
	  /* Emit the last move even if it's a no-op move, so that we
	     have something to tag REG_EQUAL notes on.  */
	  emit_insn (gen_movsi (operands[0], reg));
	  DONE;
	}
      else
	emit_insn (gen_movsi (reg, operands[1]));
      while (--i >= 0)
	emit_insn (gen_ashrsi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movsi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 16)
    emit_insn (gen_ashrsi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashrsi3_n ());
    }
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R2_REG)));
  DONE;
}")

(define_insn "ashrsi3_n"
  [(set (reg:SI R2_REG) (ashiftrt:SI (reg:SI R2_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:HI R4_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashrsi_n"
  [(set_attr "type" "sfunc")])

(define_insn "ashrsi3_c"
  [(set (reg:SI R2_REG)
	(ashiftrt:SI (reg:SI R2_REG)
		     (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 16U"
  "jsr ___ashrsi_%O0"
  [(set_attr "type" "sfunc")])

(define_insn "ashrsi3_1"
  [(set (match_operand:SI 0 "register_operand" "+r")
	(ashiftrt:SI (match_dup 0) (const_int 1)))]
  ""
  "sra %S0% rrc %R0"
  [(set_attr "length" "4")])

;; Logical Shift Right
(define_expand "lshrsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand" "")
	(lshiftrt:SI (match_operand:SI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4 || (INTVAL (operands[2]) >= 16)))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (SImode);
      i = INTVAL (operands[2]);
      if (i >= 16)
	{
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));

	  /* ??? Can't use gen_highpart here because it refuses to work on
	     SUBREGs with SUBREG_WORD != 0.  */
	  emit_insn (gen_lshrhi3 (gen_lowpart (HImode, reg),
				  operand_subword (operands[1], 0, 0, SImode),
				  GEN_INT (i - 16)));
	  emit_insn (gen_movhi (gen_highpart (HImode, reg), const0_rtx));
	  /* Emit the last move even if it's a no-op move, so that we
	     have something to tag REG_EQUAL notes on.  */
	  emit_insn (gen_movsi (operands[0], reg));
	  DONE;
	}
      else
	emit_insn (gen_movsi (reg, operands[1]));
      while (--i >= 0)
	emit_insn (gen_lshrsi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movsi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movsi (gen_rtx_REG (SImode, R2_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) < 16)
    emit_insn (gen_lshrsi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_lshrsi3_n ());
    }
  emit_insn (gen_movsi (operands[0], gen_rtx_REG (SImode, R2_REG)));
  DONE;
}")

(define_insn "lshrsi3_n"
  [(set (reg:SI R2_REG) (lshiftrt:SI (reg:SI R2_REG) (reg:HI R4_REG)))
   (clobber (reg:HI T_REG))
   (clobber (reg:HI R4_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___lshrsi_n"
  [(set_attr "type" "sfunc")])

(define_insn "lshrsi3_c"
  [(set (reg:SI R2_REG)
	(lshiftrt:SI (reg:SI R2_REG)
		     (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 16U"
  "jsr ___lshrsi_%O0"
  [(set_attr "type" "sfunc")])

(define_insn "lshrsi3_1"
  [(set (match_operand:SI 0 "register_operand" "+r")
	(lshiftrt:SI (match_dup 0) (const_int 1)))
   (clobber (reg:PQI T_REG))]
  ""
  "sr %S0% rrc %R0"
  [(set_attr "length" "4")
   (set_attr "type" "t_set")])


;; Rotate Left
(define_expand "rotlsi3"
  [(set (match_operand:SI 0 "register_operand" "")
	(rotate:SI (match_operand:SI 1 "register_operand" "")
		   (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotlsi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn_and_split "rotlsi3_1"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(rotate:SI (match_operand:SI 1 "register_operand" "0")
		   (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; add %R0,%R0% adc %S0,%S0% incc %R0
  "reload_completed"
  [(pc)]
  "
{
  rtx low = risc16_split_operand (operands[0], 2, 1);
  rtx high = risc16_split_operand (operands[0], 2, 0);

  emit_insn (gen_addhic (low, low, low));
  emit_insn (gen_adchic (high, high, high));
  emit_insn (gen_incscc (low));
  DONE;
}")

;; Rotate Right
(define_expand "rotrsi3"
  [(set (match_operand:SI 0 "register_operand" "")
	(rotatert:SI (match_operand:SI 1 "register_operand" "")
		     (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotrsi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn_and_split "rotrsi3_1"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(rotatert:SI (match_operand:SI 1 "register_operand" "0")
		     (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; sr %S0% rrc %R0% adc %S0,%S0% rr %S0
  "reload_completed"
  [(pc)]
  "
{
  rtx low = risc16_split_operand (operands[0], 2, 1);
  rtx high = risc16_split_operand (operands[0], 2, 0);

  emit_insn (gen_lshrhic (high));
  emit_insn (gen_rrc (low));
  emit_insn (gen_adchic (high, high, high));
  emit_insn (gen_rotrhi3_1 (high, high, const1_rtx));
  DONE;
}")

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "rrc"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(ior:HI (lshiftrt:HI (match_dup 0) (const_int 1))
		(ashift:HI (reg:HI T_REG) (const_int 15))))
   (set (reg:HI T_REG) (not:HI (and:HI (match_dup 0) (const_int 1))))]
  ""
  "rrc %0"
  [(set_attr "type" "t_set")])


;; ::::::::::::::::::::
;; ::
;; :: 64 bit Integer Shifts and Rotates
;; ::
;; ::::::::::::::::::::

;; Arithmetic Shift Left
(define_expand "ashldi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
	(ashift:DI (match_operand:DI 1 "general_operand" "")
		   (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4
	  || INTVAL (operands[2]) % 16U < 3
	  || INTVAL (operands[2]) >= 32))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (DImode);
      i = INTVAL (operands[2]) % 64U;
      switch (i >> 4)
	{
	case 0:
	  emit_insn (gen_movdi (reg, operands[1]));
	  break;
	case 1:
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 0),
				operand_subword (operands[1], 1, 0, DImode)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 1),
				operand_subword (operands[1], 2, 0, DImode)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 2),
				operand_subword (operands[1], 3, 0, DImode)));
	  emit_insn (gen_movhi (gen_lowpart (HImode, reg), const0_rtx));
	  break;
	case 2:
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_ashlsi3 (gen_rtx_SUBREG (SImode, reg, 0),
				  gen_lowpart (SImode, operands[1]),
				  GEN_INT (i - 32)));
	  emit_insn (gen_movsi (gen_lowpart (SImode, reg), const0_rtx));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	case 3:
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_ashlhi3 (gen_highpart (HImode, reg),
				  gen_lowpart (HImode, operands[1]),
				  GEN_INT (i - 48)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 1), const0_rtx));
	  emit_insn (gen_movsi (gen_lowpart (SImode, reg), const0_rtx));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	}
      i %= 16U;
      while (--i >= 0)
	emit_insn (gen_adddi3 (reg, reg, reg));
      if (reg != operands[0])
	emit_insn (gen_movdi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movdi (gen_rtx_REG (DImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_ashldi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashldi3_n ());
    }
  emit_insn (gen_movdi (operands[0], gen_rtx_REG (DImode, R0_REG)));
  DONE;
}")

(define_insn "ashldi3_n"
  [(set (reg:DI R0_REG) (ashift:DI (reg:DI R0_REG) (reg:HI R4_REG)))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashldi_n"
  [(set_attr "type" "sfunc")])

(define_insn "ashldi3_c"
  [(set (reg:DI R0_REG)
	(ashift:DI (reg:DI R0_REG) (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 32U"
  "jsr ___ashldi_%O0"
  [(set_attr "type" "sfunc")])

;; Arithmetic Shift Right
(define_expand "ashrdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
	(ashiftrt:DI (match_operand:DI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4
	  || INTVAL (operands[2]) % 16U < 3
	  || INTVAL (operands[2]) >= 32))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (DImode);
      i = INTVAL (operands[2]) % 64U;
      switch (i >> 4)
	{
	case 0:
	  emit_insn (gen_movdi (reg, operands[1]));
	  break;
	case 1:
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 3),
				operand_subword (operands[1], 2, 0, DImode)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 2),
				operand_subword (operands[1], 1, 0, DImode)));
	  emit_insn (gen_extendhisi2 (gen_rtx_SUBREG (SImode, reg, 0),
				      gen_highpart (HImode, operands[1])));
	  break;
	case 2:
	  emit_insn (gen_extendsidi2 (reg,
				      gen_rtx_SUBREG (SImode, operands[1], 0)));
	  emit_insn (gen_ashrsi3 (gen_lowpart (SImode, reg),
				  gen_lowpart (SImode, reg),
				  GEN_INT (i - 32)));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	case 3:
	  emit_insn (gen_extendhidi2 (reg,
				      gen_highpart (HImode, operands[1])));
	  emit_insn (gen_ashrhi3 (gen_lowpart (HImode, reg),
				  gen_lowpart (HImode, reg),
				  GEN_INT (i - 48)));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	}
      i %= 16U;
      while (--i >= 0)
	emit_insn (gen_ashrdi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movdi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movdi (gen_rtx_REG (DImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_ashrdi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_ashrdi3_n ());
    }
  emit_insn (gen_movdi (operands[0], gen_rtx_REG (DImode, R0_REG)));
  DONE;
}")

(define_insn "ashrdi3_n"
  [(set (reg:DI R0_REG) (ashiftrt:DI (reg:DI R0_REG) (reg:HI R4_REG)))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___ashrdi_n"
  [(set_attr "type" "sfunc")])

(define_insn "ashrdi3_c"
  [(set (reg:DI R0_REG)
	(ashiftrt:DI (reg:DI R0_REG) (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 32"
  "jsr ___ashrdi_%O0"
  [(set_attr "type" "sfunc")])

(define_insn "ashrdi3_1"
  [(set (match_operand:DI 0 "register_operand" "+r")
	(ashiftrt:DI (match_dup 0) (const_int 1)))
   (clobber (reg:HI T_REG))]
  ""
  "sra %U0% rrc %T0% rrc %S0% rrc %R0"
  [(set_attr "length" "8")
   (set_attr "type" "t_set")])

;; Logical Shift Right
(define_expand "lshrdi3"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
	(lshiftrt:DI (match_operand:DI 1 "general_operand" "")
		     (match_operand:HI 2 "general_operand" "")))]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT
      && (INTVAL (operands[2]) < 4
	  || INTVAL (operands[2]) % 16U < 3
	  || INTVAL (operands[2]) >= 32))
    {
      rtx reg;
      int i;

      reg = operands[0];
      if (GET_CODE (reg) != REG
	  || (REGNO (reg) < FIRST_PSEUDO_REGISTER
	      && ! rtx_equal_p (reg, operands[1])
	      && reg_overlap_mentioned_p (reg, operands[1])))
	reg = gen_reg_rtx (DImode);
      i = INTVAL (operands[2]) % 64U;
      switch (i >> 4)
	{
	case 0:
	  emit_insn (gen_movdi (reg, operands[1]));
	  break;
	case 1:
	  if (! rtx_equal_p (reg, operands[1]))
	    emit_insn (gen_rtx_CLOBBER (VOIDmode, reg));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 3),
				operand_subword (operands[1], 2, 0, DImode)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 2),
				operand_subword (operands[1], 1, 0, DImode)));
	  emit_insn (gen_movhi (gen_rtx_SUBREG (HImode, reg, 1),
				operand_subword (operands[1], 0, 0, DImode)));
	  emit_insn (gen_movhi (gen_highpart (HImode, reg), const0_rtx));
	  break;
	case 2:
	  emit_insn (gen_zero_extendsidi2 (reg,
					   gen_rtx_SUBREG (SImode, operands[1],
							   0)));
	  emit_insn (gen_lshrsi3 (gen_lowpart (SImode, reg),
				  gen_lowpart (SImode, reg),
				  GEN_INT (i - 32)));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	case 3:
	  emit_insn (gen_zero_extendhidi2 (reg,
					   gen_highpart (HImode, operands[1])));
	  emit_insn (gen_lshrhi3 (gen_lowpart (HImode, reg),
				  gen_lowpart (HImode, reg),
				  GEN_INT (i - 48)));
	  emit_insn (gen_movdi (operands[0], reg));
	  DONE;
	}
      i %= 16U;
      while (--i >= 0)
	emit_insn (gen_lshrdi3_1 (reg));
      if (reg != operands[0])
	emit_insn (gen_movdi (operands[0], reg));
      DONE;
    }
  emit_insn (gen_movdi (gen_rtx_REG (DImode, R0_REG), operands[1]));
  if (GET_CODE (operands[2]) == CONST_INT)
    emit_insn (gen_lshrdi3_c (operands[2]));
  else
    {
      emit_insn (gen_movhi (gen_rtx_REG (HImode, R4_REG), operands[2]));
      emit_insn (gen_lshrdi3_n ());
    }
  emit_insn (gen_movdi (operands[0], gen_rtx_REG (DImode, R0_REG)));
  DONE;
}")

(define_insn "lshrdi3_n"
  [(set (reg:DI R0_REG) (lshiftrt:DI (reg:DI R0_REG) (reg:HI R4_REG)))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___lshrdi_n"
  [(set_attr "type" "sfunc")])

(define_insn "lshrdi3_c"
  [(set (reg:DI R0_REG)
	(lshiftrt:DI (reg:DI R0_REG) (match_operand:HI 0 "const_int_operand" "n")))
   (clobber (reg:SI R4_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  "INTVAL (operands[0]) < 32U"
  "jsr ___lshrdi_%O0"
  [(set_attr "type" "sfunc")])

(define_insn "lshrdi3_1"
  [(set (match_operand:DI 0 "register_operand" "+r")
	(lshiftrt:DI (match_dup 0) (const_int 1)))
   (clobber (reg:HI T_REG))]
  ""
  "sr %U0% rrc %T0% rrc %S0% rrc %R0"
  [(set_attr "length" "8")
   (set_attr "type" "t_set")])

;; Rotate Left
(define_expand "rotldi3"
  [(set (match_operand:DI 0 "register_operand" "")
	(rotate:DI (match_operand:DI 1 "register_operand" "")
		   (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotldi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn_and_split "rotldi3_1"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(rotate:DI (match_operand:DI 1 "register_operand" "0")
		     (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; add %R0,%R0% adc %S0,%S0% adc %T0,%T0% adc %U0,%U0% incc %R0"
  "reload_completed"
  [(pc)]
  "
{
  rtx word, low = risc16_split_operand (operands[0], 4, 3);
  int i;

  emit_insn (gen_addhic (low, low, low));
  for (i = 2; i >= 0; i--)
    {
      word = risc16_split_operand (operands[0], 4, i);
      emit_insn (gen_adchic (word, word, word));
    }
  emit_insn (gen_incscc (low));
  DONE;
}")

;; Rotate Right
(define_expand "rotrdi3"
  [(set (match_operand:DI 0 "register_operand" "")
	(rotatert:DI (match_operand:DI 1 "register_operand" "")
		     (match_operand:HI 2 "const_1_operand" "")))]
  ""
  "
{
  if (! const_1_operand (operands[2], HImode))
    FAIL;
  emit_insn (gen_rotrdi3_1 (operands[0], operands[1], operands[2]));
  DONE;
}")

(define_insn_and_split "rotrdi3_1"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(rotatert:DI (match_operand:DI 1 "register_operand" "0")
		     (match_operand:HI 2 "const_1_operand" "n")))
   (clobber (reg:PQI T_REG))]
  ""
  "#" ;; sr %U0% rrc %T0% rrc %S0% rrc %R0% adc %U0,%U0% rr %U0
  "reload_completed"
  [(pc)]
  "
{
  rtx word, high = risc16_split_operand (operands[0], 2, 0);
  int i;

  emit_insn (gen_lshrhic (high));
  for (i = 1; i <= 3; i++)
    {
      word = risc16_split_operand (operands[0], 2, i);
      emit_insn (gen_rrc (word));
    }
  emit_insn (gen_adchic (high, high, high));
  emit_insn (gen_rotrhi3_1 (high, high, const1_rtx));
  DONE;
}")


;; ::::::::::::::::::::
;; ::
;; :: 16 Bit Integer Logical operations
;; ::
;; ::::::::::::::::::::

;; Logical AND, 16 bit integers
;; Don't use 'z' for reloading, because if that hard register happens 
;; to be used, we fail.  ??? Why can't reload disparage alternatives
;; where most / all registers are not available?
(define_insn "andhi3"
  [(set (match_operand:HI 0 "register_operand" "=!z,r,r")
	(and:HI (match_operand:HI 1 "register_operand" "%0,0,0")
		(match_operand:HI 2 "nonmemory_operand" "LT,r,i")))
   (clobber (match_scratch:PQI 3 "=X,X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "and %0,%R2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Inclusive OR, 16 bit integers
;; Don't use 'z' for reloading, because if that hard register happens 
;; to be used, we fail.  ??? Why can't reload disparage alternatives
;; where most / all registers are not available?
(define_insn "iorhi3"
  [(set (match_operand:HI 0 "register_operand" "=!z,r,r")
	(ior:HI (match_operand:HI 1 "register_operand" "%0,0,0")
		(match_operand:HI 2 "nonmemory_operand" "LT,r,i")))
   (clobber (match_scratch:PQI 3 "=X,X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "or %0,%R2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Inclusive OR, with result in T bit
;; Don't use 'z' for reloading, because if that hard register happens 
;; to be used, we fail.  ??? Why can't reload disparage alternatives
;; where most / all registers are not available?
(define_insn "orc"
  [(set (match_operand:HI 0 "register_operand" "=!z,r,r")
	(ior:HI (match_operand:HI 1 "register_operand" "%0,0,0")
		(match_operand:HI 2 "nonmemory_operand" "LT,r,i")))
   (set (reg:HI T_REG)
	(eq:HI (ior:HI (match_dup 1) (match_dup 2)) (const_int 0)))
   (clobber (match_scratch:PQI 3 "=X,X,X"))]
  ""
  "or %0,%R2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Exclusive OR, 16 bit integers
;; Don't use 'z' for reloading, because if that hard register happens 
;; to be used, we fail.  ??? Why can't reload disparage alternatives
;; where most / all registers are not available?
(define_insn "xorhi3"
  [(set (match_operand:HI 0 "register_operand" "=!z,r,r")
	(xor:HI (match_operand:HI 1 "register_operand" "%0,0,0")
		(match_operand:HI 2 "nonmemory_operand" "LT,r,i")))
   (clobber (match_scratch:PQI 3 "=X,X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "xor %0,%R2"
  [(set_attr "length" "2,2,4")
   (set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; One's complement, 16 bit integers

(define_insn "one_cmplhi2"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(not:HI (match_operand:HI 1 "register_operand" "0,t")))
   (clobber (match_scratch:PQI 2 "=X,X"))
   (clobber (reg:HI T_REG))]
  ""
  "@
	com %0
	sbc %0,%0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set,other")])

;; ??? combine won't do this because of the T clobber in one_cmplhi2.
(define_peephole2
  [(parallel [(set (match_operand:HI 0 "register_operand" "") (reg:HI T_REG))
	      (clobber (match_scratch:PQI 1 ""))])
   (parallel [(set (match_dup 0) (not:HI (match_dup 0)))
	      (clobber (match_scratch:PQI 3 ""))
	      (clobber (reg:HI T_REG))])]
  ""
  [(parallel [(set (match_dup 0) (not:HI (reg:HI T_REG)))
	      (clobber (match_scratch:PQI 1 ""))])]
  "")

;; combine ld rn,#0 / decc rn into sbc rn,rn
;; i.e. movhi_i & decscc into ldnott
;; used when doing zero extension & subtract of a small constant.
(define_peephole2
  [(parallel [(set (match_operand:HI 0 "register_operand" "") (const_int 0))
	      (clobber (match_scratch:PQI 1 ""))])
   (parallel [(set (match_dup 0) (plus:HI (match_dup 0) (reg:HI T_REG)))
	      (set (reg:HI T_REG)
		   (ior:HI (ne:HI (match_dup 0) (const_int 0)) (reg:HI T_REG)))
	      (clobber (match_scratch:PQI 2 "=X"))])]
  ""
  [(parallel [(set (match_dup 0) (not:HI (reg:HI T_REG)))
	      (clobber (match_dup 1))])]
  "")

;; same with addhic interspersed.
(define_peephole2
  [(parallel [(set (match_operand:HI 0 "register_operand" "") (const_int 0))
	      (clobber (match_scratch:PQI 1 ""))])
   (parallel [(set (match_operand:HI 3 "register_operand" "")
		   (plus:HI (match_dup 3)
			    (match_operand:HI 4 "immediate_operand" "")))
	      (set (reg:HI T_REG)
		   (ltu:HI (plus:HI (match_dup 3) (match_dup 4))
			   (match_dup 3)))
	      (clobber (match_scratch:PQI 5 ""))])
   (parallel [(set (match_dup 0) (plus:HI (match_dup 0) (reg:HI T_REG)))
	      (set (reg:HI T_REG)
		   (ior:HI (ne:HI (match_dup 0) (const_int 0)) (reg:HI T_REG)))
	      (clobber (match_scratch:PQI 2 "=X"))])]
  "! reg_overlap_mentioned_p (operands[0], operands[3])"
  [(parallel [(set (match_dup 3) (plus:HI (match_dup 3) (match_dup 4)))
	      (set (reg:HI T_REG)
		   (ltu:HI (plus:HI (match_dup 3) (match_dup 4))
			   (match_dup 3)))
	      (clobber (match_dup 5))])
   (parallel [(set (match_dup 0) (not:HI (reg:HI T_REG)))
	      (clobber (match_dup 1))])]
  "")

;; ::::::::::::::::::::
;; ::
;; :: 22 Bit Integer Logical operations
;; ::
;; ::::::::::::::::::::

;; Logical AND, 22 bit integers - this is used for stack adjusts,
;; generated by combine.
(define_insn_and_split "andpsi3_i"
  [(set (match_operand:PSI 0 "register_operand" "=a")
	(and:PSI (match_operand:PSI 1 "register_operand" "%0")
		 (match_operand:PSI 2 "const_int_operand" "N")))
   (clobber (reg:PQI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (AND, operands[0], operands[2]); DONE;")


;; ::::::::::::::::::::
;; ::
;; :: 32 Bit Integer Logical operations
;; ::
;; ::::::::::::::::::::

;; Logical AND, 32 bit integers
(define_insn_and_split "andsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (match_operand:SI 1 "register_operand" "%0")
		(match_operand:SI 2 "nonmemory_operand" "ri")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (AND, operands[0], operands[2]); DONE;")

;; Inclusive OR, 32 bit integers
(define_insn_and_split "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (match_operand:SI 1 "register_operand" "%0")
		(match_operand:SI 2 "nonmemory_operand" "ri")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (IOR, operands[0], operands[2]); DONE;")

;; Exclusive OR, 32 bit integers
(define_insn_and_split "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(xor:SI (match_operand:SI 1 "register_operand" "%0")
		(match_operand:SI 2 "nonmemory_operand" "ri")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (XOR, operands[0], operands[2]); DONE;")

;; One's complement, 32 bit integers

(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "=x,r")
	(not:SI (match_operand:SI 1 "register_operand" "0,0")))
   (clobber (match_scratch:PQI 2 "=X,c"))
   (clobber (match_scratch:PQI 3 "=X,c"))
   (clobber (reg:HI T_REG))]
  ""
  "com %R0% com %S0"
  [(set_attr "length" "4")])


;; ::::::::::::::::::::
;; ::
;; :: 64 Bit Integer Logical operations
;; ::
;; ::::::::::::::::::::

;; Logical AND, 64 bit integers
(define_insn_and_split "anddi3"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(and:DI (match_operand:DI 1 "register_operand" "%0")
		(match_operand:DI 2 "nonmemory_operand" "rn")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (AND, operands[0], operands[2]); DONE;")

;; Includive OR, 64 bit integers
(define_insn_and_split "iordi3"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(ior:DI (match_operand:DI 1 "register_operand" "%0")
		(match_operand:DI 2 "nonmemory_operand" "rn")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (IOR, operands[0], operands[2]); DONE;")

;; Exclusive OR, 64 bit integers
(define_insn_and_split "xordi3"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(xor:DI (match_operand:DI 1 "register_operand" "%0")
		(match_operand:DI 2 "nonmemory_operand" "rn")))
   (clobber (reg:HI T_REG))]
  ""
  "#"
  "reload_completed"
  [(pc)]
  "risc16_split_logical_operation (XOR, operands[0], operands[2]); DONE;")

;; One's complement, 64 bit integers
(define_insn "one_cmpldi2"
  [(set (match_operand:DI 0 "register_operand" "=x,r")
	(not:DI (match_operand:DI 1 "register_operand" "0,0")))
   (clobber (match_scratch:PQI 2 "=X,c"))
   (clobber (match_scratch:PQI 3 "=X,c"))
   (clobber (reg:HI T_REG))]
  ""
  "com %R0% com %S0% com %T0% com %U0"
  [(set_attr "length" "8")])


;; ::::::::::::::::::::
;; ::
;; :: Comparisons
;; ::
;; ::::::::::::::::::::

;; Note, we store the operands in the comparison insns, and use them later
;; when generating the branch or scc operation.

;; First the routines called by the machine independent part of the compiler
(define_expand "cmphi"
  [(set (reg:HI T_REG)
        (compare (match_operand:HI 0 "arith_reg_operand" "")
  		 (match_operand:HI 1 "nonmemory_operand" "")))]
  ""
  "
{
  risc16_compare_op0 = operands[0];
  risc16_compare_op1 = operands[1];
  DONE;
}")

(define_expand "cmppsi"
  [(set (reg:HI T_REG)
        (compare (match_operand:PSI 0 "register_operand" "")
  		 (match_operand:PSI 1 "nonmemory_operand" "")))]
  ""
  "
{
  risc16_compare_op0 = operands[0];
  risc16_compare_op1 = operands[1];
  DONE;
}")

(define_expand "cmpsi"
  [(set (reg:HI T_REG)
        (compare (match_operand:SI 0 "register_operand" "")
  		 (match_operand:SI 1 "nonmemory_operand" "")))]
  ""
  "
{
  risc16_compare_op0 = operands[0];
  risc16_compare_op1 = operands[1];
  DONE;
}")

;; no cmpdi expander; we rely on the machine-independent code to generate
;; them from multiple compares.
;; No cmpsf / cmpdf expanders; we rely on the machine-indenpendent code
;; to generate library calls.

;; Now, the actual comparisons, generated by the branch and/or scc operations

;; Special patterns for eq / ge so that the short immediate forms can be
;; put in delay slots.
(define_insn "cmpeqhi_t"
  [(set (reg:HI T_REG)
	(eq:HI (match_operand:HI 0 "arith_reg_operand" "r,d,r")
	       (match_operand:HI 1 "nonmemory_operand" "r,L,?i")))]
  ""
  "cmp eq,%0,%1"
  [(set_attr "length" "2,2,4")
   (set_attr "type" "t_set")])

(define_insn "cmpgehi_t"
  [(set (reg:HI T_REG)
	(ge:HI (match_operand:HI 0 "register_operand" "r,d,r")
	       (match_operand:HI 1 "nonmemory_operand" "r,J,?i")))]
  ""
  "cmp ge,%0,%1"
  [(set_attr "length" "2,2,4")
   (set_attr "type" "t_set")])

(define_insn_and_split "cmphi_t"
  [(set (reg:HI T_REG)
	(match_operator:HI 2 "risc16_greater_operator"
	  [(match_operand:HI 0 "register_operand" "r,r")
           (match_operand:HI 1 "nonmemory_operand" "r,?i")]))]
  ""
  "cmp %o2,%0,%R1"
;; Canonicalisation in loop replaces (GE x const) with (GT x <const-1>) .
;; Reverse this transformation after reload if this makes the insn smaller.
  "reload_completed
   && GET_CODE (operands[2]) == GT
   && true_regnum (operands[0]) < FIRST_ADDR_REG
   && GET_CODE (operands[1]) == CONST_INT
   && CONST_OK_FOR_J (INTVAL (operands[1]) + 1)"
  [(const_int 0)]
  "emit_insn (gen_cmpgehi_t (operands[0], plus_constant (operands[1], 1)));
   DONE;"
  [(set_attr "length" "2,4")
   (set_attr "type" "t_set")])

(define_insn "*cmpeqpsi_t"
  [(set (reg:HI T_REG)
	(eq:HI (match_operand:PSI 0 "register_operand" "a,a,?d,?d")
	       (match_operand:PSI 1 "nonmemory_operand" "a,?i,d,?n")))]
  ""
  "@
	cmp eq,%0,%1
	cmp eq,%0,%1
	#
	#"
  [(set_attr "length" "2,4,6,10")
   (set_attr "type" "t_set")])

(define_insn "*cmpeqsi_t"
  [(set (reg:HI T_REG)
	(eq:HI (match_operand:SI 0 "register_operand" "r,r")
	       (match_operand:SI 1 "nonmemory_operand" "r,?n")))]
  ""
  "#"
  [(set_attr "length" "6,10")])

(define_split
  [(set (reg:HI T_REG)
	(eq:HI (match_operand 0 "register_operand" "")
	       (match_operand 1 "nonmemory_operand" "")))]
  "reload_completed
   && HARD_REGNO_NREGS (true_regnum (operands[0]), GET_MODE (operands[0])) == 2"
  [(const_int 0)]
  "
{
  rtx label = gen_label_rtx ();
  rtx jump;

  emit_insn (gen_cmpeqhi_t (gen_highpart (HImode, operands[0]),
			    gen_highpart (HImode, operands[1])));
  jump = emit_jump_insn (gen_branch_false (label));
  JUMP_LABEL (jump) = label;
  emit_insn (gen_cmpeqhi_t (gen_lowpart (HImode, operands[0]),
			    gen_lowpart (HImode, operands[1])));
  emit_label (label);
  DONE;
}")

(define_insn_and_split "cmpsi_t"
  [(set (reg:HI T_REG)
	(match_operator:HI 2 "risc16_greater_operator"
	  [(match_operand:SI 0 "register_operand" "r,r")
           (match_operand:SI 1 "nonmemory_operand" "r,???i")]))]
  ""
  "#"
  "reload_completed"
  [(const_int 0)]
  "
{
  rtx label = gen_label_rtx ();
  rtx label2 = gen_label_rtx ();
  rtx jump;

  enum rtx_code code2 = GET_CODE (operands[2]);

  emit_insn (gen_cmpeqhi_t (risc16_split_operand (operands[0], 2, 0),
			    risc16_split_operand (operands[1], 2, 0)));
  jump = emit_jump_insn (gen_branch_true (label));
  JUMP_LABEL (jump) = label;
  emit_insn (gen_cmphi_t (risc16_split_operand (operands[0], 2, 0),
			  risc16_split_operand (operands[1], 2, 0),
			  operands[2]));
  jump = emit_jump_insn (gen_jump (label2));
  JUMP_LABEL (jump) = label2;
  emit_label (label);
  if (code2 == GT)
    code2 = GTU;
  else if (code2 == GE)
    code2 = GEU;
  emit_insn (gen_cmphi_t (risc16_split_operand (operands[0], 2, 1),
			  risc16_split_operand (operands[1], 2, 1),
			  gen_rtx_fmt_ee (code2, HImode, pc_rtx, pc_rtx)));
  emit_label (label2);
  DONE;
}")


;; ::::::::::::::::::::
;; ::
;; :: Branches
;; ::
;; ::::::::::::::::::::

;; Define_expands called by the machine independent part of the compiler.
;; Each of these named patterns
;; must be present, and they cannot be amalgamated into one pattern.

(define_expand "beq" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (EQ);
  emit_jump_insn (gen_branch_true (operands[0]));
  DONE;
}")

(define_expand "bne" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (EQ);
  emit_jump_insn (gen_branch_false (operands[0]));
  DONE;
}")

(define_expand "blt" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GE);
  emit_jump_insn (gen_branch_false (operands[0]));
  DONE;
}")

(define_expand "ble" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GT);
  emit_jump_insn (gen_branch_false (operands[0]));
  DONE;
}")

(define_expand "bgt" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GT);
  emit_jump_insn (gen_branch_true (operands[0]));
  DONE;
}")

(define_expand "bge" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GE);
  emit_jump_insn (gen_branch_true (operands[0]));
  DONE;
}")

(define_expand "bltu" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GEU);
  emit_jump_insn (gen_branch_false (operands[0]));
  DONE;
}")

(define_expand "bleu" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GTU);
  emit_jump_insn (gen_branch_false (operands[0]));
  DONE;
}")

(define_expand "bgtu" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GTU);
  emit_jump_insn (gen_branch_true (operands[0]));
  DONE;
}")

(define_expand "bgeu" [(match_operand 0 "" "")] ""
  "
{
  risc16_emit_compare (GEU);
  emit_jump_insn (gen_branch_true (operands[0]));
  DONE;
}")

;; Actual branches.  There is no need to allow for the (label_ref) and the
;; (pc) to be swapped, since we only have integer conditions, and these
;; are reversible by changing the comparison operator.
   
(define_insn "branch_true"
  [(set (pc) (if_then_else (ne (reg:HI T_REG) (const_int 0))
                           (label_ref (match_operand 0 "" ""))
                           (pc)))]
  ""
  "* return output_branch (\"brt\\0jpt\", insn, operands, \"brfd\");"
  [(set_attr "type" "cbranch")])


(define_insn "branch_false"
  [(set (pc) (if_then_else (eq (reg:HI T_REG) (const_int 0))
                           (label_ref (match_operand 0 "" ""))
                           (pc)))]
  ""
  "* return output_branch (\"brf\\0jpf\", insn, operands, \"brtd\");"
  [(set_attr "type" "cbranch")])

;; bnzd has only a short offset range, and no absolute address variant.
;; So we remain undecided till insn output if Z0 / Z1 is used for the
;; branch decision or the current value of r6/r7.
;; If the bnzd is out ot range, we use sub rn,#1 / brt(d) or jpt .
;; The label is operand zero so that output_branch and the length
;; attributes work without hassle.
(define_insn "bnzd"
  [(set (pc) (if_then_else (if_then_else (unspec [(pc)] 1)
					 (match_operand:HI 1
					  "register_operand" "+j,k")
					 (eq:HI (match_operand:PQI 2
						 "register_operand" "+v,w")
						(const_int 0)))
			   (label_ref (match_operand 0 "" ""))
			   (pc)))
   (set (match_dup 1)
	(plus:HI (match_dup 1) (const_int -1)))
   (set (match_dup 2)
	(eq:PQI (plus:HI (match_dup 1) (const_int -1)) (const_int 0)))
   (clobber (reg:HI T_REG))]
  ""
  "*
{
  int seq_len = dbr_sequence_length ();

  if (get_attr_length (insn) - seq_len >= 4)
    {
      output_asm_insn (\"sub #1,%1\", operands);
      return output_branch (\"brt\\0jpt\", insn, operands, \"brfd\");
    }
  else
    {
      output_asm_insn (\"bnzd %1,%O0\", operands);
      return seq_len ? \"\" : \"nop\";
    }
}"
  [(set_attr "type" "bnzd")
   (set_attr "needs_delay_slot" "yes")])

;; Implement decrement_and_branch_until_zero with bnzd for r6/r7 if Z0/Z1
;; is available.  Other cases are handled with sub rn,#1 / brt(d).
;; The decision what to use and the transformation into the more explict
;; form is done in MACHINE_DEPENDENT_REORG; it also has to set Z0 / Z1
;; live in the loop.
(define_expand "decrement_and_branch_until_zero"
  [(parallel [(set (pc)
                   (if_then_else
                    (ge (plus:HI (match_operand:HI 0 "general_operand" "")
                                 (const_int -1))
                        (const_int 0))
                    (label_ref (match_operand 1 "" ""))
                    (pc)))
              (set (match_dup 0)
                   (plus:HI (match_dup 0)
                            (const_int -1)))])]
  ""
  "emit_insn (gen_dbra_i (operands[0], operands[1])); DONE;")

(define_insn_and_split "dbra_i"
  [(set (pc)
        (if_then_else
          (ge (plus:HI (match_operand:HI 0 "register_operand" "+j,k,?x")
                       (const_int -1))
              (const_int 0))
          (label_ref (match_operand 1 "" ""))
          (pc)))
   (set (match_dup 0)
        (plus:HI (match_dup 0)
                 (const_int -1)))
   (clobber (match_scratch:PQI 2 "=v,w,X"))
   (clobber (reg:HI T_REG))]
  "find_reg_note (insn, REG_NONNEG, 0)"
  "#"
  "reload_completed
   && (! LOOP_COUNT_REGNUM_P (true_regnum (operands[0])) || split_dbra_p)"
  [(pc)]
  "
{
  rtx sub = gen_addhi3 (operands[0], operands[0], GEN_INT (-1));
  XVECEXP (sub, 0, 1) = operands[2];
  emit_insn (sub);
  emit_jump_insn (gen_branch_true (operands[1]));
}")


;; ::::::::::::::::::::
;; ::
;; :: Set flag operations
;; ::
;; ::::::::::::::::::::

;; Define_expands called by the machine independent part of the compiler
;; to allocate a new comparison register

(define_expand "seq" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (EQ);
  emit_insn (gen_ldnott (operands[0]));
  emit_insn (gen_one_cmplhi2 (operands[0], operands[0]));
  DONE;
}")

(define_expand "sne" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (EQ);
  emit_insn (gen_ldnott (operands[0]));
  DONE;
}")

(define_expand "slt" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (GE);
  emit_insn (gen_ldnott (operands[0]));
  DONE;
}")

(define_expand "sle" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (GT);
  emit_insn (gen_ldnott (operands[0]));
  DONE;
}")

(define_expand "sgt" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_scc (GT, GE, operands);
  DONE;
}")

(define_expand "sge" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_scc (GE, GT, operands);
  DONE;
}")

(define_expand "sltu" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (GEU);
  emit_insn (gen_ldnott (operands[0]));
  DONE;
}")

(define_expand "sleu" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_compare (GTU);
  emit_insn (gen_ldnott (operands[0]));
  DONE;
}")

(define_expand "sgtu" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_scc (GTU, GEU, operands);
  DONE;
}")

(define_expand "sgeu" [(match_operand:HI 0 "arith_reg_operand" "")] ""
  "
{
  risc16_emit_scc (GEU, GTU, operands);
  DONE;
}")

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "incscc"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(minus:HI (match_dup 0) (reg:HI T_REG)))
   (set (reg:HI T_REG)
	(and:HI (eq:HI (match_dup 0) (const_int -1)) (reg:HI T_REG)))
   (clobber (match_scratch:PQI 1 "=X"))]
  ""
  "incc %0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; Note that the value for a 'true' T_REG is minus 1.
(define_insn "decscc"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(plus:HI (match_dup 0) (reg:HI T_REG)))
   (set (reg:HI T_REG)
	(ior:HI (ne:HI (match_dup 0) (const_int 0)) (reg:HI T_REG)))
   (clobber (match_scratch:PQI 1 "=X"))]
  ""
  "decc %0"
  [(set_attr "loop_count_set" "yes")
   (set_attr "type" "t_set")])

;; ::::::::::::::::::::
;; ::
;; :: Call and branch instructions
;; ::
;; ::::::::::::::::::::

;; Subroutine call instruction returning no value.  Operand 0 is the function
;; to call; operand 1 is the number of bytes of arguments pushed (in mode
;; `SImode', except it is normally a `const_int'); operand 2 is the number of
;; registers used as operands.

;; On most machines, operand 2 is not actually stored into the RTL pattern.  It
;; is supplied for the sake of some RISC machines which need to put this
;; information into the assembler code; they can put it in the RTL instead of
;; operand 1.

(define_expand "call"
  [(parallel [(call (match_operand:HI 0 "jump_mem_operand" "")
		    (match_operand 1 "" ""))
	      (clobber (reg:PSI LINK_REG))])]
  ""
  "")

(define_insn "call_i"
  [(call (mem:HI (match_operand:PSI 0 "jump_operand" "R,i,a"))
	 (match_operand 1 "" ""))
   (clobber (reg:PSI LINK_REG))]
  ""
  "* return (which_alternative == 2
	     ? \"jsr %0\" : output_branch (\"bsr\\0jsr\", insn, operands, 0));"
  [(set_attr "length" "2,4,2")
   (set_attr "needs_delay_slot" "yes,no,no")
   (set_attr "in_delay_slot" "no")])

;; Subroutine call instruction returning a value.  Operand 0 is the hard
;; register in which the value is returned.  There are three more operands, the
;; same as the three operands of the `call' instruction (but with numbers
;; increased by one).

;; Subroutines that return `BLKmode' objects use the `call' insn.

(define_expand "call_value"
  [(parallel [(set (match_operand 0 "register_operand" "")
		   (call (match_operand:HI 1 "jump_mem_operand" "")
			 (match_operand 2 "" "")))
	      (clobber (reg:PSI LINK_REG))])]
  ""
  "")

(define_insn "call_value_i"
  [(set (match_operand 0 "register_operand" "=r,r,r")
	(call (mem:HI (match_operand:PSI 1 "jump_operand" "R,i,a"))
	      (match_operand 2 "" "")))
   (clobber (reg:PSI LINK_REG))]
  ""
  "* return (which_alternative == 2
	     ? \"jsr %1\"
	     : output_branch (\"bsr\\0jsr\", insn, operands + 1, 0));"
  [(set_attr "length" "2,4,2")
   (set_attr "needs_delay_slot" "yes,no,no")
   (set_attr "in_delay_slot" "no")])

;; Subroutine return
(define_expand "return"
  [(return)]
  "reload_completed && ! risc16_need_epilogue ()"
  "")

(define_insn "*return_i"
  [(return)]
  "reload_completed"
  "* return dbr_sequence_length () ? \"retd\" : \"jmp a14\";"
  [(set_attr "return_delay_slot" "yes")])

(define_insn "*return_true"
  [(if_then_else (ne (reg:HI T_REG) (const_int 0)) (return) (const_int 0))]
  "reload_completed"
  "jpt a14")

(define_insn "*return_false"
  [(if_then_else (eq (reg:HI T_REG) (const_int 0)) (return) (const_int 0))]
  "reload_completed"
  "jpf a14")

;; Normal unconditional jump
(define_insn "jump"
  [(set (pc) (label_ref (match_operand 0 "" "")))]
  ""
  "* return output_branch (\"bra\\0jmp\", insn, operands, 0);"
  [(set_attr "type" "branch")])

;; Indirect jump through a register
;; This implementation also allows constants so that reload can use
;; equivalences.
(define_insn "indirect_jump"
  [(set (pc) (mem:HI (match_operand:PSI 0 "jump_operand" "R,i,a")))]
  ""
  "* return (which_alternative == 2
	     ? \"jmp %0\" : output_branch (\"bra\\0jmp\", insn, operands, 0));"
  [(set_attr "type" "branch")
   (set_attr "length" "2,4,2")
   (set_attr "optional_delay_slot" "yes,no,no")])

;; case instruction for switch statements.

;; Operand 0 is index
;; operand 1 is the minimum bound
;; operand 2 is the maximum bound - minimum bound
;; operand 3 is CODE_LABEL for the table;
;; operand 4 is the CODE_LABEL to go to if index out of range.

(define_expand "casesi"
  [(match_operand:SI 0 "register_operand" "")
   (match_operand:SI 1 "immediate_operand" "")
   (match_operand:SI 2 "immediate_operand" "")
   (match_operand 3 "" "") (match_operand 4 "" "")]
  ""
  "
{
  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) <= 0x10000)
    {
      /* gen_highpart panics when it sees a SUBREG with a non-zero SUBREG_WORD,
	 which can happen when switching on a long long value - then we
	 get the SImode lowpart of a DImode reg.  So use operand_subword.  */
      rtx reg0 = gen_rtx_REG (HImode, R0_REG);
      rtx regx = gen_reg_rtx (HImode);
      emit_move_insn (reg0, operand_subword (operands[0], 1, 0, SImode));
      if (GET_CODE (operands[1]) == CONST_INT
	  && INTVAL (operands[1]) >= 0
	  && INTVAL (operands[1]) + INTVAL (operands[2]) <= 0xffff)
	{
	  if (INTVAL (operands[1]) + INTVAL (operands[2]) == 0xffff)
	    emit_insn (gen_subhic (reg0, reg0, operands[1]));
	  else
	    {
	      emit_insn (gen_subhi3 (reg0, reg0,
				     GEN_INT (INTVAL (operands[1])
					      + INTVAL (operands[2]) + 1)));
	      emit_insn (gen_addhic (reg0, reg0, GEN_INT (INTVAL (operands[2])
							  + 1)));
	    }
	  emit_insn (gen_ldnott (regx));
	  emit_insn (gen_orc (regx, regx,
			      operand_subword (operands[0], 0, 0, SImode)));
	  emit_jump_insn (gen_branch_false (operands[4]));
	}
      else
	{
	  emit_move_insn (regx, operand_subword (operands[0], 0, 0, SImode));
	  emit_insn (gen_subhic (reg0, reg0,
				 operand_subword (operands[1], 1, 0, SImode)));
	  emit_insn (gen_sbchic (regx, regx,
				 operand_subword (operands[1], 0, 0, SImode)));
	  emit_insn (gen_orc (regx, regx, regx));
	  emit_jump_insn (gen_branch_false (operands[4]));
	  emit_insn (gen_cmphi_t (reg0, operands[2],
				  gen_rtx_fmt_ee (GTU, HImode, reg0, reg0)));
	  emit_jump_insn (gen_branch_true (operands[4]));
	}
      emit_jump_insn (gen_casehi_worker (operands[3]));
    }
  else
    {
      rtx reg0 = gen_rtx_REG (SImode, R0_REG);
      emit_move_insn (reg0, operands[0]);
      emit_insn (gen_subsi3 (reg0, reg0, operands[1]));
      emit_insn (gen_cmpsi_t (reg0, operands[2],
			      gen_rtx_fmt_ee (GTU, HImode, reg0, reg0)));
      emit_jump_insn (gen_branch_true (operands[4]));
      emit_jump_insn (gen_casesi_worker (operands[3]));
    }
  DONE;
}")

(define_insn "casesi_worker"
  [(set (pc) (unspec [(reg:SI R0_REG) (label_ref (match_operand 0 "" ""))] 0))
   (clobber (reg:SI R0_REG))
   (clobber (reg:PSI LINK_REG))
   (clobber (reg:PQI T_REG))]
  ""
  "jsr ___casesi"
  [(set_attr "length" "4")])

(define_insn "casehi_worker"
  [(set (pc) (unspec [(reg:HI R0_REG) (label_ref (match_operand 0 "" ""))] 0))
   (clobber (reg:HI R0_REG))
   (clobber (reg:PSI LINK_REG))
   (clobber (reg:PQI T_REG))]
  ""
  "*
{
  rtx diff_vec = PATTERN (next_real_insn (operands[0]));

  if (GET_CODE (diff_vec) != ADDR_DIFF_VEC)
    abort ();

  switch (GET_MODE (diff_vec))
    {
    case SImode:
      if (XVECLEN (diff_vec, 1) <= 0x8000)
	return \"jsr	___casephi22\";
      else
	return \"jsr	___casehi22\";
    case HImode:
      return \"jsr	___casehi16\";
    case PHImode:
      return \"bsrd	0f\;add	r0,r0\";
    case QImode:
      return \"jsr	___casehi8\";
    default:
      abort ();
    }
}"
;; In the PHImode case, there are four bytes of code at the end of the
;; table.  Since we can't adjust the length of the ADDR_DIFF_VEC,
;; pretend these four bytes belong to this insn.
;; To avoid oscillation, we also say these bytes are needed for HImode.
  [(set (attr "length")
	(cond [(ne (symbol_ref "GET_MODE (PATTERN (next_real_insn (operands[0]))) == PHImode") (const_int 0))
	       (const_int 8)
	       (ne (symbol_ref "GET_MODE (PATTERN (next_real_insn (operands[0]))) == HImode") (const_int 0))
	       (const_int 8)
	       (ne (match_dup 0) (match_dup 0))
	       (const_int 4)] (const_int 4)))])

;; This is an alternate implementation of casehi, which makes the branch
;; unnecesary, and thus the dispatch slightly quicker.
;; However, for all but the PHImode variant, in the case that the value
;; is out of bounds, the decision will be delayed, and there will be an
;; out-of-bounds access to program memory.
(define_insn "casehi_worker2"
  [(set (pc)
	(if_then_else (ne (reg:HI T_REG) (const_int 0))
		      (unspec [(match_operand:HI 1 "register_operand" "+a")
			       (label_ref (match_operand 0 "" ""))] 0)
		      (label_ref (match_operand 2 "" ""))))
   (clobber (match_dup 1))
   (clobber (reg:PSI LINK_REG))]
  "! optimize_size"
  "*
{
  rtx diff_vec = PATTERN (next_real_insn (operands[0]));

  if (GET_CODE (diff_vec) != ADDR_DIFF_VEC)
    abort ();

  switch (GET_MODE (diff_vec))
    {
    case SImode:
      if (XVECLEN (diff_vec, 1) < 0x80000)
	return \"ld a14,pc\;add a14,#20\;add a%N1,a%N1\;add a14,%1\;add a14,%1\;ldc %1,@a14\;add a14,#2\;ldc r14,@a14\;ld e14,%1\;jpt a14\;jmp %O2\";
      else
	return \"ld a14,pc\;add a14,#22\;add a14,%1\;add a14,%1\;add a14,%1\;add a14,%1\;ldc %1,@a14\;add a14,#2\;ldc r14,@a14\;ld e14,%1\;add a14,%1\;jpt a14\;jmp %O2\";
    case HImode:
      return \"ld a14,pc\;add a14,#14\;add a14,%1\;add a14,%1\;ldc %1,@a14\;add a14,%1\;jpt a14\;jmp %O2\";
    case PHImode:
      return \"bsrd	0f\;add	a%N1,a%N1\";
    default:
      abort ();
    }
}"
  [(set_attr "length" "4")])

;; ??? There is no provision for a casehi expander, so a switch on an int
;; will be expanded as sign extension & casesi.  Prune this down.
(define_peephole2
  [(set (reg:HI T_REG)
	(ge:HI (match_operand:HI 0 "register_operand" "") (const_int 0)))
   (parallel [(set (match_operand:HI 1 "register_operand" "")
		   (not:HI (reg:HI T_REG)))
	      (clobber (match_scratch:PQI 2 ""))])
   (parallel [(set (reg:HI R0_REG) (match_dup 0))
	      (clobber (match_scratch:PQI 3 ""))])
   (parallel [(set (reg:HI R0_REG)
		   (plus:HI (reg:HI R0_REG)
			    (match_operand:HI 4 "const_int_operand" "")))
	      (clobber (match_scratch:PQI 5 ""))
	      (clobber (reg:HI T_REG))])
   (parallel [(set (reg:HI R0_REG)
		   (plus:HI (reg:HI R0_REG)
			    (match_operand:HI 6 "const_int_operand" "")))
	      (set (reg:HI T_REG)
		   (ltu:HI (plus:HI (reg:HI R0_REG) (match_dup 6))
			   (reg:HI R0_REG)))
	      (clobber (match_scratch:PQI 7 ""))])
   (parallel [(set (match_operand:HI 8 "register_operand" "")
		   (not:HI (reg:HI T_REG)))
	      (clobber (match_scratch:PQI 9 ""))])
   (parallel [(set (match_dup 8) (ior:HI (match_dup 8) (match_dup 1)))
	      (set (reg:HI T_REG)
		   (eq:HI (ior:HI (match_dup 8) (match_dup 1)) (const_int 0)))
	      (clobber (match_scratch:PQI 10 ""))])]
  "peep2_reg_dead_p (7, operands[1])
   && peep2_reg_dead_p (7, operands[8])
   && ! (-1 - INTVAL (operands[4]) & 0x8000)
   && ! (-1 + INTVAL (operands[6]) & 0x8000)"
  [
   (parallel [(set (reg:HI R0_REG) (match_dup 0))
	      (clobber (match_dup 3))])
   (parallel [(set (reg:HI R0_REG) (plus:HI (reg:HI R0_REG) (match_dup 4)))
	      (clobber (match_dup 5))
	      (clobber (reg:HI T_REG))])
   (parallel [(set (reg:HI R0_REG) (plus:HI (reg:HI R0_REG) (match_dup 6)))
	      (set (reg:HI T_REG)
		   (ltu:HI (plus:HI (reg:HI R0_REG) (match_dup 6))
			   (reg:HI R0_REG)))
	      (clobber (match_dup 7))])]
  "")

;; Instruction to jump to a variable address.  This is a low-level capability
;; which can be used to implement a dispatch table when there is no `casesi'
;; pattern.  Either the 'casesi' pattern or the 'tablejump' pattern, or both,
;; MUST be present in this file.

;; This pattern requires two operands: the address or offset, and a label which
;; should immediately precede the jump table.  If the macro
;; `CASE_VECTOR_PC_RELATIVE' is defined then the first operand is an offset
;; which counts from the address of the table; otherwise, it is an absolute
;; address to jump to.  In either case, the first operand has mode `Pmode'.

;; The `tablejump' insn is always the last insn before the jump table it uses.
;; Its assembler code normally has no need to use the second operand, but you
;; should incorporate it in the RTL pattern so that the jump optimizer will not
;; delete the table as unreachable code.

(define_insn "tablejump"
  [(set (pc) (match_operand:SI 0 "register_operand" "r"))
   (use (label_ref (match_operand 1 "" "")))]
  "! CASE_VECTOR_PC_RELATIVE"
  "jmp %0"
  [(set_attr "length" "4")])


;; ::::::::::::::::::::
;; ::
;; :: Prologue and Epilogue instructions
;; ::
;; ::::::::::::::::::::

;; Called after register allocation to add any instructions needed for the
;; prologue.  Using a prologue insn is favored compared to putting all of the
;; instructions in the FUNCTION_PROLOGUE macro, since it allows the scheduler
;; to intermix instructions with the saves of the caller saved registers.  In
;; some cases, it might be necessary to emit a barrier instruction as the last
;; insn to prevent such scheduling.
(define_expand "prologue"
  [(const_int 1)]
  ""
  "
{
  risc16_expand_prologue ();
  DONE;
}")

;; Called after register allocation to add any instructions needed for the
;; epilogue.  Using a epilogue insn is favored compared to putting all of the
;; instructions in the FUNCTION_EPILOGUE macro, since it allows the scheduler
;; to intermix instructions with the restires of the caller saved registers.
;; In some cases, it might be necessary to emit a barrier instruction as the
;; first insn to prevent such scheduling.
(define_expand "epilogue"
  [(const_int 2)]
  ""
  "
{
  risc16_expand_epilogue ();
  DONE;
}")

(define_insn "push_rr"
  [(set (mem:HI (reg:PSI SP_REG)) (match_operand:HI 0 "register_operand" "d,b"))
   (set (mem:HI (plus:PSI (reg:PSI SP_REG) (const_int -2)))
	(match_operand:HI 1 "register_operand" "d,b"))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int -4)))]
  ""
  "push %0,%1")

(define_insn "push_aa"
  [(set (mem:PSI (reg:PSI SP_REG)) (match_operand:PSI 0 "register_operand" "b"))
   (set (mem:PSI (plus:PSI (reg:PSI SP_REG) (const_int -4)))
	(match_operand:PSI 1 "register_operand" "b"))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int -8)))]
  ""
  "push %0,%1")

(define_insn "push_r"
  [(set (mem:HI (reg:PSI SP_REG)) (match_operand:HI 0 "register_operand" "d,b"))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int -2)))]
  ""
  "push %0")

(define_insn "push_a"
  [(set (mem:PSI (reg:PSI SP_REG)) (match_operand:PSI 0 "register_operand" "b"))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int -4)))]
  ""
  "push %0")

(define_insn "pop_rr"
  [(set (match_operand:HI 0 "register_operand" "=d,b")
	(mem:HI (plus:PSI (reg:PSI SP_REG) (const_int 2))))
   (set (match_operand:HI 1 "register_operand" "=d,b")
	(mem:HI (plus:PSI (reg:PSI SP_REG) (const_int 4))))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int 4)))]
  ""
  "pop %0,%1"
  [(set_attr "type" "pop")])

(define_insn "pop_aa"
  [(set (match_operand:PSI 0 "register_operand" "=b")
	(mem:PSI (plus:PSI (reg:PSI SP_REG) (const_int 4))))
   (set (match_operand:PSI 1 "register_operand" "=b")
	(mem:PSI (plus:PSI (reg:PSI SP_REG) (const_int 8))))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int 8)))]
  ""
  "pop %0,%1"
  [(set_attr "type" "pop")])

(define_insn "pop_r"
  [(set (match_operand:HI 0 "register_operand" "=d,b")
	(mem:HI (plus:PSI (reg:PSI SP_REG) (const_int 2))))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int 2)))]
  ""
  "pop %0"
  [(set_attr "type" "pop")])

(define_insn "pop_a"
  [(set (match_operand:PSI 0 "register_operand" "=b")
	(mem:PSI (plus:PSI (reg:PSI SP_REG) (const_int 4))))
   (set (reg:PSI SP_REG) (plus:PSI (reg:PSI SP_REG) (const_int 4)))]
  ""
  "pop %0"
  [(set_attr "type" "pop")])

;; ::::::::::::::::::::
;; ::
;; :: Miscellaneous instructions
;; ::
;; ::::::::::::::::::::

;; No operation, needed in case the user uses -g but not -O.
(define_insn "nop"
  [(const_int 0)]
  ""
  "nop")

;; Pseudo instruction that prevents the scheduler from moving code above this
;; point.
(define_insn "blockage"
  [(unspec_volatile [(const_int 0)] 0)]
  ""
  ""
  [(set_attr "length" "0")])

;; Trap instruction to allow usage of the __builtin_trap function
(define_insn "trap"
  [(trap_if (const_int 1) (const_int 0))]
  ""
;; ??? is this the right software interrupt to call? 62 to choose from...
  "swi #50")

(define_insn "dect"
  [(set (reg:HI T_REG)
        (eq:HI (match_operand:HI 0 "register_operand" "+x,j,k") (const_int 1)))
   (set (match_dup 0) (plus:HI (match_dup 0) (const_int -1)))
   (clobber (match_scratch:PQI 1 "=X,v,w"))]
  ""
  "dt   %0"
  [(set_attr "type" "t_set")])

;; String/block move insns.

;; Note that argument passing of structures uses expand_block_move.  We
;; must not generate a call to memcpy for that, because that needs arguments
;; on the stack, which overwrite any arguments that have already been placed
;; there to be arguments to the same function as the structure is.

(define_expand "movstrhi"
  [(parallel [(set (mem:BLK (match_operand:BLK 0 "" ""))
                   (mem:BLK (match_operand:BLK 1 "" "")))
              (use (match_operand:SI 2 "nonmemory_operand" ""))
              (use (match_operand:SI 3 "immediate_operand" ""))])]
  ""
  "
{
  /* Check alignment.  */
  if (INTVAL (operands[3]) < 2)
    FAIL;
  /* Check that size is aligned.
     ??? This also disallows dynamic arrays.  Unfortunately, we don't have
     alignment information for these.  */
  if (GET_CODE (operands[2]) != CONST_INT || INTVAL (operands[2]) & 1
      /* _memcpy_6789 is really better for larger sizes.  */
      || INTVAL (operands[2]) > 40)
    FAIL;
  force_into (XEXP (operands[0], 0), gen_rtx_REG (PSImode, A8_REG));
  force_into (XEXP (operands[1], 0), gen_rtx_REG (PSImode, A9_REG));
  force_into (operands[2], gen_rtx_REG (HImode, R0_REG));
  emit_insn (gen_movstrhi_i ());
  DONE;
}")

(define_insn "movstrhi_i"
  [(set (mem:BLK (reg:PSI A8_REG))
	(mem:BLK (reg:PSI A9_REG)))
   (use (reg:HI R0_REG))
   (use (const_int 16))
   (clobber (reg:SI R2_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI A9_REG))
   (clobber (reg:PSI A10_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))]
  ""
  "jsr ___movstrhi"
  [(set_attr "type" "sfunc")])

(define_expand "movstrsi"
  [(parallel [(set (mem:BLK (match_operand:BLK 0 "" ""))
                   (mem:BLK (match_operand:BLK 1 "" "")))
              (use (match_operand:SI 2 "nonmemory_operand" ""))
              (use (match_operand:SI 3 "immediate_operand" ""))])]
  ""
  "
{
  /* Check alignment.  */
  if (INTVAL (operands[3]) < 2)
    FAIL;
  force_into (XEXP (operands[0], 0), gen_rtx_REG (PSImode, A8_REG));
  force_into (XEXP (operands[1], 0), gen_rtx_REG (PSImode, A9_REG));
  force_into (operands[2], gen_rtx_REG (SImode, R6_REG));
  emit_insn (gen_movstrsi_i ());
  DONE;
}")

(define_insn "movstrsi_i"
  [(set (mem:BLK (reg:PSI A8_REG))
	(mem:BLK (reg:PSI A9_REG)))
   (use (reg:SI R6_REG))
   (use (const_int 16))
   (clobber (reg:SI R0_REG))
   (clobber (reg:HI R2_REG))
   (clobber (reg:SI R6_REG))
   (clobber (reg:PSI A8_REG))
   (clobber (reg:PSI A9_REG))
   (clobber (reg:PSI A10_REG))
   (clobber (reg:HI T_REG))
   (clobber (reg:PSI LINK_REG))
   (clobber (reg:PQI Z0_REG))
   (clobber (reg:PQI Z1_REG))]
  ""
  "jsr ___memcpy_6789"
  [(set_attr "type" "sfunc")])
